package top.daxian1218.dust.db;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import top.daxian1218.dust.commons.data.HDataRow;
import top.daxian1218.dust.commons.data.HDataTable;
import top.daxian1218.dust.commons.generic.Preconditions;
import top.daxian1218.dust.commons.util.BeanUtil;
import top.daxian1218.dust.commons.util.ClassUtil;
import top.daxian1218.dust.db.base.*;
import top.daxian1218.dust.db.sql.DefaultDataAdapter;
import top.daxian1218.dust.db.sql.IDataSource;
import top.daxian1218.dust.db.sql.command.HTableSqlCommand;
import top.daxian1218.dust.db.sql.pool.DataSourcePool;
import top.daxian1218.dust.db.sql.pool.IDataSourceTemplate;
import top.daxian1218.dust.db.support.DataBaseFactory;

import java.io.Closeable;
import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;


/**
 * 数据库管理
 * @author hst
 */
public class DbManager {
    static Logger logger = LoggerFactory.getLogger(DbManager.class);

    private static final int ENABLE_STATUS = 1;
    private static final String DB_ACCESS = "dust_db_access ";
    private Thread syncThread;
    private boolean isStop;
    private final List<AdapterHolder> adapterContainer = new ArrayList<>();
    private IDataSourceTemplate dataSourceTemplate;
    private List<DbAccess> localDbList = new ArrayList<>();
    private String remoteConfigDb;

    public DbManager(IDataSourceTemplate dataSourceTemplate) {
        Preconditions.checkNotNull(dataSourceTemplate);
        this.dataSourceTemplate = dataSourceTemplate;

        registerAdapter(DbTypes.Mysql, DbCluster.ReadWrite, DefaultDataAdapter.class);
        registerAdapter(DbTypes.PG, DbCluster.ReadWrite, DefaultDataAdapter.class);
        registerAdapter(DbTypes.Mssql, DbCluster.ReadWrite, DefaultDataAdapter.class);
        registerAdapter(DbTypes.Oracle, DbCluster.ReadWrite, DefaultDataAdapter.class);
        registerAdapter(DbTypes.ClickHouse, DbCluster.ReadWrite, DefaultDataAdapter.class);
        registerAdapter(DbTypes.Highgo, DbCluster.ReadWrite, DefaultDataAdapter.class);
    }

    public synchronized void loadConfig() {
        for (DbAccess db : localDbList) {
            if (StringUtils.isEmpty(db.getName())) {
                throw new DustDbException("数据源未执行名称");
            }

            addDataSource(db);
        }

        if (!StringUtils.isEmpty(this.remoteConfigDb)) {
            IDataAdapter adapter = null;
            try {
                adapter = new DbAdapterManager(this).getAdapter(remoteConfigDb, true);
                HDataTable dt = adapter.query(getQueryDbAccessCommand(adapter.getDbType()));
                for (HDataRow row : dt.getRows()) {
                    DbAccess db = ClassUtil.mapToObject(row.toMap(), DbAccess.class);
                    addDataSource(db);
                }

            } catch (DustDbException e) {
                logger.debug("db access remote fail", e);
            } finally {
                if (adapter != null) {
                    adapter.closeQuiet();
                }
            }
        }
    }



    private void addDataSource(DbAccess db) {
        if (isRdb(db.getDbType())) {
            IDataSource ds = DataSourcePool.Default.get(db.getName());
            if (ds != null && !ds.getDbAccess().equals(db)) {
                if (ds instanceof Closeable) {
                    try {
                        ((Closeable) ds).close();
                    } catch (IOException e) {
                        logger.debug("关闭数据源异常", e);
                    }
                }
                DataSourcePool.Default.set(db.getName(), null);
                return;
            }

            // 再次校验， 如果存在数据源则跳过创建
            ds = DataSourcePool.Default.get(db.getName());
            if (ds != null) {
                return;
            }

            try {
                DataSourcePool.Default.set(db.getName(), dataSourceTemplate.createDataSource(db));
            } catch (SQLException e) {
                logger.info("db:" + db.getName() + " can not open");
            }
        }
    }

    private HDataCommand getQueryDbAccessCommand(String dbType) {
        if (isRdb(dbType)) {
            return new HTableSqlCommand(DB_ACCESS);
        }

        return new HDataCommand(DB_ACCESS);
    }

    public void registerAdapter(String dbType, DbCluster cluster, Class<? extends IDataAdapter> adapterClass) {
        AdapterHolder holder = new AdapterHolder();
        holder.setDbType(dbType);
        holder.setCluster(cluster);
        holder.setAdapterClass(adapterClass);

        //同一数据库，仅允许一个适配器
        for (int i = adapterContainer.size() - 1; i >= 0; i--)
        {
            if (StringUtils.equals(adapterContainer.get(i).getDbType(), dbType) &&
                    adapterContainer.get(i).getCluster() == cluster)
            {
                adapterContainer.remove(i);
                adapterContainer.add(i, holder);
                return;
            }
        }

        adapterContainer.add(holder);
    }

    public IDataAdapter newAdapter(String name) throws DustDbException {

        name = StringUtils.isEmpty(name) ? DbConstants.DEFAULT_DB_NAME : name;

        IDataSource ds = DataSourcePool.Default.get(name);
        if (ds == null)
        {
            throw new DustDbException("没有找到指定的数据源");
        }

        DbAccess dbAccess = ds.getDbAccess();
        for (int i = adapterContainer.size() - 1; i >= 0; i--) {
            AdapterHolder holder = adapterContainer.get(i);
            if (support(dbAccess, holder)) {
                try {
                    String dbOperationsClass = dbAccess.getDbOperationsClass();
                    IDbOperations dbOperations = null;
                    if (!StringUtils.isEmpty(dbOperationsClass))
                    {
                        dbOperations = (IDbOperations) BeanUtil.getBean(dbOperationsClass);
                        if (dbOperations == null)
                        {
                            throw new DustDbException("db operations bean not found named of " + dbOperationsClass);
                        }
                    }
                    else
                    {
                        dbOperations = DataBaseFactory.create(dbAccess.getDbType());
                    }

                    IDataAdapter adapter = holder.adapterClass.newInstance();
                    if (adapter != null) {
                        adapter.init(ds, dbOperations);
                        return adapter;
                    }
                } catch (InstantiationException | IllegalAccessException e) {
                    logger.debug("新建适配器异常", e);
                }
            }
        }

        throw new DustDbException("adapter not found");
    }

    private boolean support(DbAccess dbAccess, AdapterHolder holder) {
        return StringUtils.equals(dbAccess.getDbType(), holder.dbType);
    }

    private boolean isRdb(String dbType) {
        switch (dbType) {
            case DbTypes.ClickHouse:
            case DbTypes.PG:
            case DbTypes.Mssql:
            case DbTypes.Mysql:
            case DbTypes.Oracle:
            case DbTypes.Highgo:
            case DbTypes.Sqlite: {
                return true;
            }
            default:
                return false;
        }
    }

    public void startSync() {
        if (syncThread == null || !syncThread.isAlive()) {
            syncThread = new Thread(new Runnable() {
                @Override
                public void run() {
                    while (!isStop) {
                        loadConfig();
                        try {
                            Thread.sleep(5 * 60 * 1000);
                        } catch (InterruptedException e) {
                            logger.info("sync thread interrupt");
                        }
                    }
                }
            });
            syncThread.setDaemon(true);
            syncThread.start();
        }
    }

    public void stopSync() {
        if (syncThread != null) {
            isStop = true;
            syncThread.interrupt();
            try {
                syncThread.join();
            } catch (InterruptedException ex) {
                logger.debug("sync thread interrupt fail", ex);
                logger.info("sync thread interrupt fail");
            } finally {
                syncThread = null;
            }
        }
    }

    public List<DbAccess> getLocalDbList() {
        return localDbList;
    }

    public String getRemoteConfigDb() {
        return remoteConfigDb;
    }

    public void setRemoteConfigDb(String remoteConfigDb) {
        this.remoteConfigDb = remoteConfigDb;
    }

    public static class AdapterHolder {
        String dbType;
        DbCluster cluster;
        Class<? extends IDataAdapter> adapterClass;

        public String getDbType() {
            return dbType;
        }

        public void setDbType(String dbType) {
            this.dbType = dbType;
        }

        public DbCluster getCluster() {
            return cluster;
        }

        public void setCluster(DbCluster dbCluster) {
            this.cluster = dbCluster;
        }

        public Class<? extends IDataAdapter> getAdapterClass() {
            return adapterClass;
        }

        public void setAdapterClass(Class<? extends IDataAdapter> adapterClass) {
            this.adapterClass = adapterClass;
        }
    }
}
