package top.daxian1218.dust.db.base;

import org.apache.commons.lang3.StringUtils;
import top.daxian1218.dust.commons.data.HDataColumn;
import top.daxian1218.dust.commons.data.HDataRecord;
import top.daxian1218.dust.commons.data.HDataTable;
import top.daxian1218.dust.commons.generic.Preconditions;
import top.daxian1218.dust.commons.util.TypeUtil;
import top.daxian1218.dust.db.DustDbException;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.regex.Pattern;

/**
 * 数据执行命令
 *
 * @author hst
 */
public class HDataCommand {
    private String commandText;
    private final static Pattern PATTERNS = Pattern.compile("[_A-Za-z0-9]*");
    protected final List<HDataRecord> records = new ArrayList<>();

    protected final List<HDataColumn> fields = new ArrayList<>();

    private HDataRecord parameters;
    private Integer pageSize = -1;
    private Integer pageIndex = 0;
    private Integer totalRows = -1;
    private Object tag;

    protected final String PrefixIndexKey = "@";

    public HDataCommand(String text) {
        commandText = StringUtils.isEmpty(text) ? StringUtils.EMPTY : text;
        parameters = new HDataRecord();
        records.add(parameters);
    }

    public HDataCommand() {
        this(null);
    }

    public Object getTag() {
        return tag;
    }

    public void setTag(Object tag) {
        this.tag = tag;
    }

    public Integer getPageSize() {
        return pageSize;
    }

    public void setPageSize(Integer pageSize) {
        this.pageSize = pageSize;
    }

    public Integer getPageIndex() {
        return pageIndex;
    }

    public void setPageIndex(Integer pageIndex) {
        this.pageIndex = pageIndex;
    }

    public Integer getTotalRows() {
        return totalRows;
    }

    public void setTotalRows(Integer totalRows) {
        this.totalRows = totalRows;
    }

    public List<HDataRecord> getParameterList() {
        return Collections.unmodifiableList(records);
    }

    public List<HDataColumn> getFields() {
        return fields;
    }

    public HDataRecord getParameters() {
        return parameters;
    }

    public HDataRecord getParameters(int index) {
        if (index < 0 || index >= records.size()) {
            return null;
        }

        return records.get(index);
    }

    private void setParameters(HDataRecord ps) {
        parameters = ps;
    }

    public String getCommandText() {
        return commandText;
    }

    public String getExecuteText() {
        return commandText;
    }

    public void iterate(Function<HDataCommand, Boolean> it) {
        Preconditions.checkNotNull(it);

        if (records.size() == 0) {
            it.apply(this);
            return;
        }

        for (HDataRecord record : records) {
            setParameters(record);
            if (!it.apply(this)) {
                break;
            }
        }
    }

    public void applyBatch() {
        if (parameters.isEmpty()) {
            return;
        }
        parameters = new HDataRecord();
        records.add(parameters);
    }


    /**
     * 获取相应的Sql参数值
     * 默认指向参数集合列表的最后一个集合，如果需要修改其他的结合，请通过getParametersList来修改
     *
     * @param key
     * @return
     */
    public Object getParameter(String key) {
        if (parameters.isEmpty()) {
            return null;
        }
        return parameters.getValue(key);
    }

    public boolean hasParameter() {
        return !parameters.isEmpty();
    }


    /**
     * 添加Sql参数，Sql语句中以:key的形式来存放参数
     *
     * @param key
     * @param value
     * @return
     */
    public HDataCommand setParameter(String key, Object value) throws DustDbException {
        if (!key.startsWith(PrefixIndexKey) && !PATTERNS.matcher(key).matches()) {
            throw new DustDbException("parameter key only allow to be underscore,letter,number");
        }
        parameters.setValue(key, value);
        return this;
    }

    public HDataCommand appendParameter(Object value) {
        String key = PrefixIndexKey +  parameters.size();
        setParameter(key, value);
        return this;
    }

    public HDataCommand appendParameters(Map<String, Object> ps) {
        HDataRecord newRow = new HDataRecord(ps);
        return appendParameters(newRow);
    }

    public HDataCommand appendParameters(HDataRecord ps) {
        if (ps.isEmpty()) {
            return this;
        }

        int len = parameters.size();

        ps.forEachKey(key ->
        {
            String parameterName = key;
            if (key.startsWith(PrefixIndexKey)) {
                int intKey = TypeUtil.toInteger(key.substring(1));
                parameterName = String.format("%s%d", PrefixIndexKey, len + intKey);
            }

            setParameter(parameterName, ps.getValue(key));
        });

        return this;
    }

    public HDataTable query(IDataAdapter adapter) throws DustDbException {
        Preconditions.checkNotNull(adapter);

        return adapter.query(this);
    }

    public UpdateResult execute(IDataAdapter adapter) throws DustDbException {
        Preconditions.checkNotNull(adapter);

        return adapter.update(this);
    }
}
