package top.daxian1218.dust.db.base;

import top.daxian1218.dust.commons.data.HDataTable;
import top.daxian1218.dust.db.DustDbException;
import top.daxian1218.dust.db.sql.IDataSource;

import java.sql.ResultSet;
import java.util.function.Consumer;

/**
 * 适配器管理连接的创建和销毁，操作面向的会话或者客户端
 * @author hst
 */
public interface IDataAdapter {

    /**
     * 根据数据库接入配置，初始化适配器
     * @param dbAccess
     * @throws DustDbException
     */
    void init(IDataSource dataSource, IDbOperations dbOperations);

    /**
     * 获取适配的数据库类型
     * @return
     */
    String getDbType();

    /**
     * 判断适配的数据库是否已打开
     * @return
     */
    boolean isOpen();

    /**
     * 打开连接
     * <p>对于已打开的连接，该方法无效；对于已关闭的连接可新开连接</p>
     * <p>Note: 有些情况下，连接可能因为超时或者网略原因断掉，此时如果调用open连接时，仍然无效。只有显式的调用了{@link #close()}才会新创建连接</p>     *
     * @throws DustDbException
     */
    void open() throws DustDbException;

    HDataTable query(HDataCommand cmd) throws DustDbException;

    /**
     * 查询，用于对性能敏感的情况，避免数据搬运到HDataTable后，再一次复制
     * @param cmd
     * @param handler
     * @throws DustDbException
     */
    void query(HDataCommand cmd, Consumer<ResultSet> handler) throws DustDbException;

    UpdateResult update(HDataCommand cmd) throws DustDbException;

    /**
     * 提交操作，面向的是支持事务的数据库。如果不支持事务默认通过
     * @throws DustDbException
     */
    void commit() throws DustDbException;


    /**
     * 回滚操作, 面向的是支持事务的数据库。如果不支持事务默认报出异常
     * @throws DustDbException
     */
    void rollback() throws DustDbException;


    /**
     * 关闭连接
     * 关闭的连接可通过open重新打开时新的连接
     * @throws DustDbException
     */
    void close() throws DustDbException;

    /**
     * {@link #rollback()}
     */
    void rollbackQuiet();


    /**
     * {@link #commit()}
     */
    void commitAndCloseQuiet();

    /**
     * {@link #close()}
     */
    void closeQuiet();

    /**
     * 销毁适配器
     * <p>销毁的适配器不可再用</p>
     *
     */
    void destroy();

    boolean isAbandon();

    void setAbandon(boolean abandon);

}
