package top.daxian1218.dust.db.dict;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import org.apache.commons.lang3.StringUtils;
import top.daxian1218.dust.commons.data.HDataRow;
import top.daxian1218.dust.commons.data.HDataTable;
import top.daxian1218.dust.commons.generic.IDisposable;
import top.daxian1218.dust.commons.generic.Preconditions;
import top.daxian1218.dust.commons.log.LoggerHelper;
import top.daxian1218.dust.db.DictGlobalConfig;
import top.daxian1218.dust.db.DustDbException;
import top.daxian1218.dust.db.base.IDataAdapter;
import top.daxian1218.dust.db.dict.condition.DictCondition;
import top.daxian1218.dust.db.dict.event.*;
import top.daxian1218.dust.db.dict.express.Expression;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.function.Consumer;
import java.util.function.Function;

import static top.daxian1218.dust.commons.generic.Preconditions.checkNotNull;

/**
 * 字典对象
 *
 * @author hst
 */
public class DictObj implements IDisposable {
    private Integer currentIndex = -1;
    private final List<DictObjRow> deletedRows = new ArrayList<>();
    private final List<DictObjTable> relationTables = new ArrayList<>();
    private long id;
    private String key;
    private String code;
    private String appCode;
    private String name;
    private String tableName;
    /**
     * 固定查询条件，每次查询都会携带
     */
    private DictCondition fixCondition;

    /**
     * 故障查询条件，通常用于多次查询直到收到数据
     */
    private final List<DictCondition> pollyConditions = new ArrayList<>(0);

    /**
     * 故障类型，0-Retry 重试
     */
    private int pollyType = 0;

    /**
     * 普通查询条件，每次查询后都会重置
     */
    private final List<DictCondition> conditions = new ArrayList<>(0);
    private final List<OrderBy> orderBys = new ArrayList<>(0);
    private String orderBySql;
    private String fixConditionSql;
    private String conditionSql;
    private final List<Error> errors = new ArrayList<>();
    private final DictObjColumnCollection columns;
    private final DictObjRowCollection rows;
    private DictPropertyCollection extendedProperties = new DictPropertyCollection();
    private IVariableHandler variableHandler;
    private final PageInfo pageInfo = new PageInfo();
    // 是否只读，只读对象无法保存数据
    private boolean readonly;

    /**
     * 临时属性，用于标识search不需要返回数据
     */
    private boolean noData;

    // 主从关系集合
    private final List<DictObjRelation> relations = new ArrayList<>();
    // 附加信息
    private Object state;
    private Object queryParams;

    HDataTable dataTable = new HDataTable();

    public DictObj() {
        rows = new DictObjRowCollection(this);
        columns = new DictObjColumnCollection(this);
    }

    public DictObj(String code, String appCode) {
        this();
        this.code = code;
        this.appCode = appCode;
        init();
    }

    public void init() {
        if (StringUtils.isEmpty(code) && id <= 0) {
            throw new DustDictException("Code is empty or id is not positive");
        }

        if (DictGlobalConfig.Instance.getObjContainer() == null) {
            throw new DustDictException("object container is null");
        }

        DictGlobalConfig.Instance.getObjContainer().load(this);
    }

    private EventHandler<ValidateEventArgs> validateEvent;
    private EventHandler<CurrentChangedEventArgs> currentChangedEvent;
    private EventHandler<CollectionChangedEventArgs> collectionChangedEvent;
    private EventHandler<RowEventArgs> rowDeletedEvent;
    private EventHandler<RowEventArgs> rowRemovedEvent;
    private EventHandler<ValueChangedEventArgs> valueChangedEvent;
    private EventHandler<ColumnsChangedEventArgs> columnsChangedEvent;


    public DictObjRowCollection getRows() {
        return rows;
    }

    public DictObjColumnCollection getColumns() {
        return columns;
    }

    public boolean isNoData() {
        return noData;
    }

    public void setNoData(boolean noData) {
        this.noData = noData;
    }

    public long getId() {
        return id;
    }

    protected void setId(long id) {
        this.id = id;
    }

    public String getCode() {
        return code;
    }

    protected void setCode(String code) {
        this.code = code;
    }

    public String getAppCode() {
        return appCode;
    }

    protected void setAppCode(String appCode) {
        this.appCode = appCode;
    }

    public String getKey() {
        return key;
    }

    public void setKey(String key) {
        this.key = key;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public String getTableName() {
        return tableName;
    }

    public void setTableName(String tableName) {
        this.tableName = tableName;
    }

    public PageInfo getPageInfo() {
        return pageInfo;
    }

    public List<DictObjRelation> getRelations() {
        return relations;
    }

    /**
     * 获取故障查询条件集合
     * @return
     */
    public List<DictCondition> getPollyConditions() {
        return pollyConditions;
    }

    /**
     * 是否只读，只读对象无法保存数据
     *
     * @return
     */
    public boolean isReadonly() {
        return readonly;
    }

    public void setReadonly(boolean readonly) {
        this.readonly = readonly;
    }

    public Object getState() {
        return state;
    }

    public void setState(Object state) {
        this.state = state;
    }

    public Object getQueryParams() {
        return queryParams;
    }

    public void setQueryParams(Object queryParams) {
        this.queryParams = queryParams;
    }

    public DictCondition getFixCondition() {
        return fixCondition;
    }

    public String getFixConditionSql() {
        return fixConditionSql;
    }

    public void setFixConditionSql(String fixConditionSql) {
        this.fixConditionSql = fixConditionSql;
    }

    public void setFixCondition(DictCondition fixCondition) {
        this.fixCondition = fixCondition;
    }

    public List<DictCondition> getConditions() {
        return conditions;
    }

    public List<OrderBy> getOrderBys() {
        return orderBys;
    }

    public String getOrderBySql() {
        return orderBySql;
    }

    public void setOrderBySql(String orderBySql) {
        this.orderBySql = orderBySql;
    }

    public String getConditionSql() {
        return conditionSql;
    }

    public void setConditionSql(String conditionSql) {
        this.conditionSql = conditionSql;
    }

    public List<Error> getErrors() {
        return errors;
    }

    public DictPropertyCollection getExtendedProperties() {
        return extendedProperties;
    }

    public void setExtendedProperties(DictPropertyCollection extendedProperties) {
        this.extendedProperties = extendedProperties;
    }

    public String getPropertyEx(String key) {
        return this.extendedProperties.get(key);
    }

    public IVariableHandler getVariableHandler() {
        return Optional.ofNullable(variableHandler).orElse(DictGlobalConfig.Instance.getVariableHandler());
    }

    public void setVariableHandler(IVariableHandler variableHandler) {
        this.variableHandler = variableHandler;
    }

    public DictObjRow getFirstOrDefaultRow() {
        if (isEmpty()) {
            return null;
        }

        return rows.get(0);
    }

    public boolean isEmpty() {
        return rows.size() == 0;
    }

    public DictObjTable getTable() {
        DictObjTable tb = new DictObjTable(tableName, code);
        tb.setPrimary(true);
        return tb;
    }

    public DictObjTable getTable(String alias) {
        return relationTables.stream().filter(t -> StringUtils.equals(t.getTableAlias(), alias)).findFirst().get();
    }

    public List<DictObjTable> getRelationTables() {
        return this.relationTables;
    }

    public boolean hasDeletedRow() {
        return deletedRows.size() > 0;
    }

    public Integer getCurrentIndex() {
        return currentIndex;
    }

    public void setCurrentIndex(int value) {
        if (value == currentIndex) {
            return;
        }

        int oldIndex = currentIndex;
        if (value < 0 || value >= rows.size()) {
            currentIndex = -1;
        } else {
            currentIndex = value;
        }

        if (oldIndex != currentIndex) {
            onCurrentChanged(oldIndex, currentIndex);
        }
    }

    public DictObjRow getCurrentRow() {
        return rows.get(currentIndex);
    }

    public void setCurrentRow(DictObjRow value) {
        if (value == null) {
            currentIndex = -1;
            return;
        }

        currentIndex = rows.indexOf(value);
    }

    public DictObj bindRelation(DictObj detail, String masterKeys, String detailKeys) {
        Preconditions.checkRequired(detail, "detail");
        Preconditions.checkRequired(masterKeys, "masterKeys");
        Preconditions.checkRequired(detailKeys, "detailKeys");

        String[] mKeys = masterKeys.split(",");
        String[] dKeys = detailKeys.split(",");
        if (mKeys.length != dKeys.length) {
            return this;
        }

        DictObjRelation relation = new DictObjRelation(this);
        relation.setDetail(detail);
        relation.getMasterKeys().addAll(Arrays.asList(mKeys));
        relation.getDetailKeys().addAll(Arrays.asList(dKeys));

        this.relations.add(relation);
        return this;
    }

    /**
     * 校验
     *
     * @return
     */
    public boolean validate() {
        errors.clear();
        for (DictObjRow row : this.rows) {
            if (row.validate()) {
                continue;
            }

            errors.addAll(row.getErrors());
            return false;
        }

        return onValidate();
    }

    protected boolean onValidate() {
        if (this.validateEvent != null) {
            ValidateEventArgs args = new ValidateEventArgs(this, null);
            validateEvent.apply(this, args);
            return args.isValid();
        }

        return true;
    }

    public void loadData(HDataTable table) {
        this.clear();
        for (HDataRow row : table.getRows()) {
            DictObjRow r = new DictObjRow(this, true);
            r.loadData(row);
            rows.load(r);
        }
        dataTable = table;
        syncDataTable();
        this.onCollectionChanged(ActionType.RESET);
    }

    public DictObjRow moveTo(int idx) {
        if (this.rows.size() > idx) {
            int oldIndex = this.currentIndex;
            this.currentIndex = idx;
            this.onCurrentChanged(oldIndex, idx);
            return this.rows.get(idx);
        }

        return null;
    }

    protected void onCurrentChanged(int oldIndex, int newIndex) {
        if (this.currentChangedEvent != null) {
            this.currentChangedEvent.apply(this, new CurrentChangedEventArgs(this, oldIndex, newIndex));
        }
    }

    protected void onValueChanged(DictObjRow row, DictObjColumn column, Object newValue, Object oldValue) {
        if (valueChangedEvent != null) {
            valueChangedEvent.apply(this, new ValueChangedEventArgs(this, row, column, newValue, oldValue));
        }
    }

    protected void onColumnsChanged(ActionType actionType, DictObjColumn column) {
        if (columnsChangedEvent != null) {
            columnsChangedEvent.apply(this, new ColumnsChangedEventArgs(column, actionType));
        }
    }

    protected void onCollectionChanged(ActionType actionType) {
        if (this.collectionChangedEvent != null) {
            this.collectionChangedEvent.apply(this, new CollectionChangedEventArgs(this, actionType));
        }
    }

    protected void onRowDeleted(DictObjRow row) {
        if (rowDeletedEvent != null) {
            this.rowDeletedEvent.apply(this, new RowEventArgs(this, row));
        }
    }


    protected void onRowRemoved(DictObjRow row) {
        if (rowRemovedEvent != null) {
            this.rowRemovedEvent.apply(this, new RowEventArgs(this, row));
        }
    }

    public DictObjColumn getColumn(int index) {
        if (index >= this.columns.size()) {
            throw new IllegalArgumentException("index");
        }

        return this.columns.get(index);
    }

    public DictObjColumn getColumn(String name) {
        return columns.get(name);
    }

    public DictObjRow getRow(int index) {
        return this.rows.get(index);
    }

    private void syncDataTable() {
        List<Integer> removedCols = new ArrayList<>();
        dataTable.getColumns().forEach(col -> {
            if (col.getName().endsWith(DictConstant.REF_COL_SUF)) {
                String name = StringUtils.stripEnd(col.getName(), DictConstant.REF_COL_SUF);
                DictObjColumn column = getColumn(name);
                if (column != null) {
                    col.setName(column.getColumnName() + DictConstant.REF_COL_SUF);
                } else {
                    removedCols.add(col.getIndex());
                }
                return;
            }
            DictObjColumn column = getColumn(col.getName());
            if (column != null) {
                col.setName(column.getColumnName());
            } else {
                removedCols.add(col.getIndex());
            }
        });

        removedCols.sort((a, b) -> b - a);
        for (Integer index : removedCols) {
            dataTable.removeColumnAt(index);
        }
    }

    public JSONObject toGridJSON() {
        syncDataTable();
        dataTable.getColumns().forEach(col -> {
            if (col.getName().endsWith("$")) {
                String name = StringUtils.trim("$");
                getColumn(name);
            }
        });
        return dataTable.toGridJSON();
    }

    public JSONObject toPageGridJSON() {
        JSONObject pageJson = new JSONObject();
        pageJson.put("size", this.getPageInfo().getTotalRows());
        pageJson.put("page", this.toGridJSON());
        return pageJson;
    }

    public JSONArray toDataJSON() {
        JSONArray jsonArray = new JSONArray();
        for (DictObjRow r : rows) {
            jsonArray.add(r.toJSON());
        }
        return jsonArray;
    }

    public JSONObject toPageDataJSON() {
        JSONObject pageJson = new JSONObject();
        pageJson.put("size", this.getPageInfo().getTotalRows());
        pageJson.put("page", this.toDataJSON());
        return pageJson;
    }

    public DictObjRow newRow() {
        return new DictObjRow(this);
    }

    public DictObjRow addRow() {
        return addRow(null);
    }

    public DictObjRow addRow(DictObjRow dataObjRow) {
        if (dataObjRow == null) {
            dataObjRow = newRow();
        }

        if (dataObjRow.getDictObj() != this) {
            throw new DustDictException("无效的行");
        }

        if (dataObjRow.getState() == RowState.DELETED) {
            throw new DustDictException("已删除的行，不允许再使用");
        }

        if (dataObjRow.getState() == RowState.DETACHED) {
            this.rows.add(dataObjRow);
        }
        dataObjRow.setState(RowState.ADDED);
        return dataObjRow;
    }

    /**
     * 业务修改，强修改，会把已存在的行强制置为修改状态；如果不存在则以修改行加入dataObj
     *
     * @param row
     */
    public void modifyRow(DictObjRow row) {
        if (this.rows.contains(row)) {
            row.setState(RowState.MODIFIED);
            this.onCollectionChanged(ActionType.REPLACE);
            return;
        }

        if (row.getState() == RowState.DETACHED && row.getDictObj() == this) {
            rows.add(row, RowState.MODIFIED);
        }
    }

    public void clear() {
        this.rows.clear();
        this.deletedRows.clear();
        this.dataTable.clear();
        this.onCollectionChanged(ActionType.RESET);
    }

    /**
     * 业务方法，如果dataObjRow已在DataObj则移除行的同时做删除操作，如果dataObjRow不再DataObj则做业务标记操作，做删除操作
     *
     * @param row
     */
    public void deleteRow(DictObjRow row) {
        checkNotNull(row);
        if (!this.rows.contains(row)) {
            row.acceptChanges();
            row.setState(RowState.DELETED);
            this.deletedRows.add(row);
            return;
        }

        this.rows.remove(row);
        row.rejectChanges();
        row.setState(RowState.DELETED);
        deletedRows.add(row);
        this.onRowDeleted(row);
    }

    /**
     * 还原修改，注意，行的顺序不会还原
     * 增加行移除，修改行还原，删除行从末尾补上
     */
    public void rejectChanges() {
        int len = this.rows.size();
        for (int i = len; i > 0; i--) {
            DictObjRow row = this.rows.get(i - 1);
            if (row.getState() == RowState.ADDED) {
                this.rows.remove(row);
            }

            if (row.getState() == RowState.MODIFIED) {
                row.rejectChanges();
            }
        }

        for (DictObjRow row : this.deletedRows) {
            row.setState(RowState.UNCHANGED);
            this.rows.add(row);
        }

        this.deletedRows.clear();

        this.onCollectionChanged(ActionType.RESET);
    }

    /**
     * 接受修改，通常用于save操作
     */
    public void acceptChanges() {
        for (DictObjRow row : this.rows) {
            row.acceptChanges();
        }

        this.deletedRows.clear();
        this.onCollectionChanged(ActionType.RESET);
    }

    public void removeRow(DictObjRow row) {
        rows.remove(row);
    }

    public void deleteAll() {
        while (rows.size() > 0) {
            deleteRow(rows.get(0));
        }
    }

    public void deleteRows(Function<DictObjRow, Boolean> func) {
        Preconditions.checkNotNull(func);
        for (DictObjRow row : this.rows) {
            Boolean ret = func.apply(row);
            if (ret != null && ret) {
                deleteRow(row);
            }
        }
    }

    public List<DictObjRow> getChanges() {
        List<DictObjRow> tmpRows = new ArrayList<>();

        for (DictObjRow r : rows) {
            if (r.getState() != RowState.UNCHANGED) {
                tmpRows.add(r);
            }
        }

        tmpRows.addAll(deletedRows);
        return tmpRows;
    }

    public void iterateUpdateRows(Function<DictObjRow, Boolean> iterator) {
        Preconditions.checkNotNull(iterator);
        for (DictObjRow r : rows) {
            if (r.getState() == RowState.MODIFIED) {
                if (!iterator.apply(r)) {
                    return;
                }
            }
        }
    }

    public void iterateInsertRows(Function<DictObjRow, Boolean> iterator) {
        Preconditions.checkNotNull(iterator);
        for (DictObjRow r : rows) {
            if (r.getState() == RowState.ADDED) {
                if (!iterator.apply(r)) {
                    return;
                }
            }
        }
    }

    public void forEachChanges(Consumer<DictObjRow> action) {
        for (DictObjRow r : rows) {
            if (r.getState() != RowState.UNCHANGED && r.getState() != RowState.DETACHED) {
                action.accept(r);
            }
        }

        for (DictObjRow r : deletedRows) {
            action.accept(r);
        }
    }

    public List<DictObjRow> filter(String expr) {
        Expression expression = new Expression(this, expr);
        return expression.execute();
    }

    public void filter(String expr, Function<DictObjRow, Boolean> iterator) {
        checkNotNull(iterator);
        Expression expression = new Expression(this, expr);
        expression.execute(iterator);

    }

    private boolean beforeSave() {
        if (this.readonly) {
            return false;
        }
        if (getVariableHandler() != null) {
            iterateUpdateRows(row -> {
                getColumns().forEach(col -> {
                    if (!StringUtils.isEmpty(col.getDefaultValue())
                            && col.getDefaultValue().startsWith("@")) {
                        row.get(col.getColumnName()).setValue(getVariableHandler().getUpdateValue(col, row));
                    }
                });
                return true;
            });
        }

        return validate();
    }

    public boolean save() throws DustDbException {
        return save(null);
    }

    public boolean save(IDataAdapter adapter) throws DustDbException {
        if (beforeSave()) {
            SqlBuilderFactory.getBuilder(adapter).save(this, adapter, adapter == null);
            acceptChanges();
            return true;
        }

        return false;
    }

    public DictObj addCondition(DictCondition condition) {
        if (condition == null || condition.isEmpty()) {
            return this;
        }

        return addCondition(condition, true);
    }

    public DictObj addFixCondition(DictCondition condition) {
        if (condition == null || condition.isEmpty()) {
            return this;
        }

        if (this.fixCondition == null) {
            this.fixCondition = new DictCondition();
        }

        this.fixCondition.getSubConditions().add(condition);
        return this;
    }

    /**
     * 添加条件，如果append为false，则会清空已存在的条件，使用condition
     *
     * @param condition
     * @param append
     * @return
     */
    public DictObj addCondition(DictCondition condition, boolean append) {
        checkNotNull(condition);
        if (condition.getLeft() == null
                && condition.getRight() == null
                && condition.getSubConditions().size() == 0) {
            LoggerHelper.debug("无效的条件");
            return this;
        }

        if (!append) {
            conditions.clear();
        }
        this.conditions.add(condition);
        return this;
    }

    public DictObj addCondition(String columnName, Object value) {
        addCondition(new DictCondition(columnName, value));
        return this;
    }

    public DictObj addOrderBy(OrderBy orderBy) {
        Preconditions.checkNotNull(orderBy);
        Preconditions.checkNotEmpty(orderBy.getColumnName());
        this.orderBys.add(orderBy);
        return this;
    }


    public void search() throws DustDbException {
        search(null);
    }

    public void search(IDataAdapter adapter) throws DustDbException {
        if (noData) {
            this.noData = false;
            this.clear();
            return;
        }
        SqlBuilderFactory.getBuilder(adapter).search(this, adapter);
        conditions.clear();
    }

    public void search(long id) {
        search(id, null);
    }

    public void search(long id, IDataAdapter adapter) {
        List<DictCondition> tempConditions = new ArrayList<>(conditions);
        String tempSql = conditionSql;
        String tempFixSql = fixConditionSql;
        DictCondition tempFixCondition = fixCondition;
        int tempPageSize = this.pageInfo.getPageSize();

        conditions.clear();
        this.pageInfo.setPageSize(0);
        conditionSql = StringUtils.EMPTY;
        fixCondition = null;
        fixConditionSql = StringUtils.EMPTY;
        conditions.add(new DictCondition("id", id));
        try {
            SqlBuilderFactory.getBuilder(adapter).search(this, adapter);
        } finally {
            conditions.clear();
            conditions.addAll(tempConditions);
            conditionSql = tempSql;
            fixConditionSql = tempFixSql;
            fixCondition = tempFixCondition;
            this.pageInfo.setPageSize(tempPageSize);
        }
    }

    //数据交互结束

    //数据对象结构
    public DictObj parseJson(JSONObject jsonObject) {
        this.tableName = jsonObject.getString("tableName");
        if (StringUtils.isEmpty(this.tableName)) {
            throw new IllegalArgumentException("empty tableName not compatible");
        }

        this.id = jsonObject.getLong("id");
        this.name = jsonObject.getString("name");
        this.code = jsonObject.getString("code");
        this.appCode = jsonObject.getString("appCode");
        this.orderBySql = jsonObject.getString("orderBySql");
        this.fixConditionSql = jsonObject.getString("fixConditionSql");

        //cols
        this.columns.clear();
        JSONArray arrCol = jsonObject.getJSONArray("columns");
        for (int i = 0; i < arrCol.size(); i++) {
            JSONObject jsonCol = arrCol.getJSONObject(i);
            this.columns.add(DictObjColumn.parse(jsonCol));
        }

        //sub tables
        this.relationTables.clear();
        JSONArray arrTable = jsonObject.getJSONArray("relationTables");
        if (arrTable != null) {
            for (int i = 0; i < arrTable.size(); i++) {
                JSONObject jsonTable = arrTable.getJSONObject(i);
                DictObjTable tb = DictObjTable.parse(jsonTable);
                this.relationTables.add(tb);
            }
        }

        if (jsonObject.containsKey("pageInfo")) {
            JSONObject jsonPage = jsonObject.getJSONObject("pageInfo");
            this.pageInfo.setStart(null2Zero(jsonPage.getInteger("start")));
            this.pageInfo.setPageSize(null2Zero(jsonPage.getInteger("pageSize")));
        }

        //orders
        JSONArray arrOrders = jsonObject.getJSONArray("orderBy");
        if (arrOrders != null) {
            for (int i = 0; i < arrOrders.size(); i++) {
                JSONObject jsOder = arrOrders.getJSONObject(i);
                OrderBy orderBy = new OrderBy();
                orderBy.setColumnName(jsOder.getString("columnName"));
                orderBy.setOrderByType(OrderByType.valueOf(jsOder.getString("orderByType")));
                this.orderBys.add(orderBy);
            }
        }

        //fixCondition
        JSONObject jsonFixCondition = jsonObject.getJSONObject("fixCondition");
        if (jsonFixCondition != null && !jsonFixCondition.isEmpty()) {
            this.fixCondition = DictCondition.parse(jsonFixCondition);
        }


        String strExProps = jsonObject.getString("extendedProperties");
        if (!StringUtils.isEmpty(strExProps) && strExProps.startsWith("{")) {
            JSONObject exProps = JSONObject.parseObject(strExProps);
            for (String k : exProps.keySet()) {
                this.extendedProperties.add(k, exProps.getString(k));
            }
        }
        return this;
    }

    private Integer null2Zero(Integer n) {
        return n == null ? 0 : n;
    }

    JSONObject schemaJson = new JSONObject();

    public JSONObject toJson() {
        schemaJson.put("id", this.id);
        schemaJson.put("name", this.name);
        schemaJson.put("code", this.code);
        schemaJson.put("tableName", this.tableName);

        //cols
        JSONArray arrCol = new JSONArray();
        for (int i = 0; i < columns.size(); i++) {
            arrCol.add(columns.get(i).toJson());
        }
        schemaJson.put("columns", arrCol);

        //sub tables
        JSONArray arrTable = new JSONArray();
        for (DictObjTable table : relationTables) {
            arrTable.add(table.toJson());
        }
        schemaJson.put("tables", arrTable);

        schemaJson.put("orderBySql", this.orderBySql);
        schemaJson.put("fixConditionSql", this.fixConditionSql);
        schemaJson.put("pageInfo", this.pageInfo);

        //orders
        JSONArray arrOrders = new JSONArray();
        for (OrderBy orderBy : orderBys) {
            JSONObject jsOder = new JSONObject();
            jsOder.put("columnName", orderBy.getColumnName());
            jsOder.put("orderByType", orderBy.getOrderByType());
            arrOrders.add(jsOder);

        }
        schemaJson.put("orderBy", arrOrders);

        //fixCondition
        if (this.fixCondition != null) {
            schemaJson.put("fixCondition", this.fixCondition.toJson());
        }

        return schemaJson;
    }

    public void parseSql(String sql, String dbType) {
        //TODO 暂不支持，后续开发 依赖于SqlBuilder
    }

    @Override
    public void dispose() {
        this.rows.dispose();
        this.columns.dispose();
        this.relationTables.clear();
        extendedProperties.clear();
        this.validateEvent = null;
        this.rowRemovedEvent = null;
        this.rowDeletedEvent = null;
        this.collectionChangedEvent = null;
        this.currentChangedEvent = null;
        valueChangedEvent = null;
        columnsChangedEvent = null;
        variableHandler = null;
    }

    /**
     * 获取故障查询类型， 0-以此尝试，直到查询到数据
     * @return 类型
     */
    public int getPollyType() {
        return pollyType;
    }

    /**
     * 设置故障查询类型
     * @param pollyType 类型
     */
    public void setPollyType(int pollyType) {
        this.pollyType = pollyType;
    }
}
