package top.daxian1218.dust.db.dict;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import org.apache.commons.lang3.StringUtils;
import top.daxian1218.dust.commons.data.HDataRow;
import top.daxian1218.dust.commons.data.HDataTable;
import top.daxian1218.dust.commons.generic.Preconditions;
import top.daxian1218.dust.commons.util.NameUtil;
import top.daxian1218.dust.commons.util.TypeUtil;
import top.daxian1218.dust.db.DbAdapterManager;
import top.daxian1218.dust.db.DictGlobalConfig;
import top.daxian1218.dust.db.DustDbException;
import top.daxian1218.dust.db.base.DbTypes;
import top.daxian1218.dust.db.base.HDataCommand;
import top.daxian1218.dust.db.base.IDataAdapter;
import top.daxian1218.dust.db.sql.HSqlCommand;
import top.daxian1218.dust.db.sql.SqlConstants;
import top.daxian1218.dust.db.sql.command.HTableSqlCommand;

import static top.daxian1218.dust.commons.generic.Preconditions.checkNotNull;

/**
 * @author hst on 2020-8-1.
 */
public class DictObjContainer implements IDictObjContainer {
    protected final String DictObj = "dict_obj";
    protected final String DictObjColumn = "dict_obj_column";
    protected final String DictObjTable = "dict_obj_table";
    protected final DbAdapterManager dbAdapterManager;
    private static final int RETRY_MAX = 3;

    public DictObjContainer(DbAdapterManager dbAdapterManager) {
        this.dbAdapterManager = dbAdapterManager;
    }

    public boolean isSupport(String dbType) {
        return StringUtils.equals(dbType, DbTypes.Mssql)
                || StringUtils.equals(dbType, DbTypes.Oracle)
                || StringUtils.equals(dbType, DbTypes.Sqlite)
                || StringUtils.equals(dbType, DbTypes.Mysql)
                || StringUtils.equals(dbType, DbTypes.PG)
                || StringUtils.equals(dbType, DbTypes.Highgo);
    }

    protected HDataTable retryQuery(HDataCommand cmd) {
        Exception innerException = null;
        for (int i = 0; i < RETRY_MAX; i++) {

            IDataAdapter sqlAdapter = getDataAdapter();
            try {
                return sqlAdapter.query(cmd);
            } catch (Exception ex) {
                innerException = ex;
            } finally {
                sqlAdapter.destroy();
            }
        }

        throw new DustDbException("字典加载失败", "0008", innerException);
    }

    public long getObjId(String app, String code) {
        HSqlCommand sqlCmd = new HSqlCommand(SqlConstants.Select + " id " + SqlConstants.From + " " + DictObj);
        sqlCmd.appendWhere("code=:code");
        sqlCmd.setParameter("code", code);

        if (!StringUtils.isEmpty(app)) {
            sqlCmd.appendWhere("app_code=:app");
            sqlCmd.setParameter("app", app);
        }

        HDataTable dt = retryQuery(sqlCmd);
        return dt.size() > 0 ? TypeUtil.toLong(dt.getScalar()) : 0L;
    }

    public HDataTable getObjData(long id) {
        HSqlCommand masterCmd = new HTableSqlCommand(DictObj);
        masterCmd.appendWhere("id=:id");
        masterCmd.setParameter("id", id);
        return retryQuery(masterCmd);
    }

    public HDataTable getObjColumnData(long objId) {
        HSqlCommand columnCmd = new HTableSqlCommand(DictObjColumn);
        columnCmd.appendWhere("obj_id = :objId");
        columnCmd.setParameter("objId", objId);
        columnCmd.addOrder("seq_no");
        return retryQuery(columnCmd);
    }

    public HDataTable getObjTable(long objId) {
        HSqlCommand tableCmd = new HTableSqlCommand(DictObjTable);
        tableCmd.appendWhere("obj_id = :objId");
        tableCmd.setParameter("objId", objId);
        tableCmd.appendOrderString("seq_no");
        return retryQuery(tableCmd);
    }


    @Override
    public IDataAdapter getDataAdapter() {
        IDataAdapter sqlAdapter = this.dbAdapterManager.getAdapter(DictGlobalConfig.Instance.getDictDb(), true);
        if (!isSupport(sqlAdapter.getDbType())) {
            throw new DustDbException("dict Container dbType not compatible for " + sqlAdapter.getDbType());
        }

        return sqlAdapter;
    }

    @Override
    public top.daxian1218.dust.db.dict.DictObj createAsync(String app, String code) {
        DictObj obj = new DictObj();
        obj.setAppCode(app);
        obj.setCode(code);
        return load(obj);
    }

    @Override
    public top.daxian1218.dust.db.dict.DictObj createAsync(long id) {
        DictObj obj = new DictObj();
        obj.setId(id);
        return load(obj);
    }

    @Override
    public top.daxian1218.dust.db.dict.DictObj load(top.daxian1218.dust.db.dict.DictObj obj) {
        Preconditions.checkNotNull(obj);
        if (obj.getId() <= 0) {
            if (StringUtils.isEmpty(obj.getCode())) {
                throw new DustDictException("id or code of obj is not set");
            }

            if (StringUtils.isEmpty(obj.getAppCode())) {
                obj.setAppCode(DictGlobalConfig.Instance.getDefaultApp());
            }

            obj.setId(getObjId(obj.getAppCode(), obj.getCode()));
        }

        JSONObject jsonOfObj = getDictObjSchema(obj.getId());
        obj.parseJson(jsonOfObj);
        return obj;
    }

    protected JSONObject getDictObjSchema(long id) {
        HDataTable tbObj = getObjData(id);
        if (tbObj.size() <= 0) {
            throw new DustDbException("数据字典未找到-" + id);
        }

        if (tbObj.size() > 1) {
            throw new DustDbException("数据字典重复-" + id);
        }

        HDataTable tbColumn = getObjColumnData(id);
        HDataTable tbTable = getObjTable(id);
        return toJsonSchema(tbObj.get(0), tbColumn, tbTable);
    }

    private boolean isJson(String str) {
        if (StringUtils.isEmpty(str)) { return false; }
        char firstChar = str.charAt(0);
        return firstChar == '[' || firstChar == '{';
    }

    protected JSONObject toJsonSchema(HDataRow objRow, HDataTable columnData, HDataTable tableData) {
        checkNotNull(objRow);
        checkNotNull(columnData);

        JSONObject jsonOfObj = row2CamelJson(objRow);

        String strFixCondition = jsonOfObj.getString("fixCondition");
        if (isJson(strFixCondition)) {
            jsonOfObj.put("fixCondition", JSONArray.parse(strFixCondition));
        } else {
            jsonOfObj.remove("fixCondition");
            jsonOfObj.put("fixConditionSql", strFixCondition);
        }

        String strOrderBy = jsonOfObj.getString("orderBy");
        if (isJson(strOrderBy)) {
            jsonOfObj.put("orderBy", JSON.parseArray(strOrderBy));
        } else {
            jsonOfObj.remove("orderBy");
            jsonOfObj.put("orderBySql", strOrderBy);
        }

        if (columnData.size() == 0) {
            jsonOfObj.put("columns", new JSONArray());
        } else {
            JSONArray arr = new JSONArray();
            for (int i = 0; i < columnData.size(); i++) {
                HDataRow r = columnData.getRows().get(i);
                arr.add(row2CamelJson(r));
            }

            jsonOfObj.put("columns", arr);
        }

        if (tableData.size() == 0) {
            jsonOfObj.put("relationTables", new JSONArray());
        } else {
            JSONArray arr = new JSONArray();
            for (int i = 0; i < tableData.size(); i++) {
                HDataRow r = tableData.getRows().get(i);
                arr.add(row2CamelJson(r));

            }
            jsonOfObj.put("relationTables", arr);
        }

        return jsonOfObj;
    }

    private JSONObject row2CamelJson(HDataRow r) {
        JSONObject json = new JSONObject();
        r.iterator(colKey -> {
            json.put(NameUtil.underscore2Camel(colKey), r.getValue(colKey));
            return true;
        });

        return json;
    }

}
