package top.daxian1218.dust.db.dict;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.alibaba.fastjson.annotation.JSONField;
import org.apache.commons.lang3.StringUtils;
import top.daxian1218.dust.commons.data.HDataRow;
import top.daxian1218.dust.commons.generic.IDisposable;
import top.daxian1218.dust.commons.generic.Preconditions;
import top.daxian1218.dust.commons.util.ClassUtil;
import top.daxian1218.dust.commons.util.TypeUtil;

import java.math.BigDecimal;
import java.util.*;
import java.util.stream.Collectors;

import static top.daxian1218.dust.commons.generic.Preconditions.checkNotNull;


/**
 * 字典行
 *
 * @author hst
 */
public class DictObjRow implements Iterable<Object>, IDisposable {
    private HDataRow data;
    private final List<Error> errors = new ArrayList<>();
    private long id;
    private String tag;
    private RowState rowState = RowState.DETACHED;

    @JSONField(serialize = false, deserialize = false)
    private DictObj dictObj;

    public DictObjRow(DictObj obj) {
        this(obj, false);
    }

    public DictObjRow(DictObj obj, boolean ignoreDefault) {
        Preconditions.checkNotNull(obj);
        this.dictObj = obj;
        init(ignoreDefault);
    }


    public long getId() {
        return id;
    }

    protected void setId(long id) {
        this.id = id;
    }

    /**
     * 数据ID，用于区分行，不具备业务逻辑，请勿与表列ID混淆
     *
     * @return 标签
     */
    public String getTag() {
        return tag;
    }

    public void setTag(String tag) {
        this.tag = tag;
    }

    public List<Error> getErrors() {
        return errors;
    }

    public DictObj getDictObj() {
        return dictObj;
    }

    public RowState getState() {
        return rowState;
    }

    protected void setState(RowState rowState) {
        this.rowState = rowState;
    }

    private void init(boolean ignoreDefault) {
        if (!ignoreDefault) {
            data = dictObj.dataTable.addRow();
            generateDefaultValue();
        }
    }

    private void generateDefaultValue() {
        this.dictObj.getColumns().forEach(column -> {
            String defaultValue = column.getDefaultValue();
            if (!StringUtils.isEmpty(defaultValue)) {
                if (column.getDictObj().getVariableHandler() != null
                        && column.isVarValue()) {
                    setValue(column, getDictObj().getVariableHandler().getDefaultValue(column, this));
                } else {
                    setValue(column, defaultValue);
                }
            } else {
                switch (column.getDataType()) {
                    case BOOLEAN:
                        setValue(column, false);
                        break;
                    case INT:
                    case NUMBER:
                    case CURRENCY:
                    case UBIGINT:
                        setValue(column, 0);
                        break;
                    case STRING:
                    case TEXT:
                        setValue(column, StringUtils.EMPTY);
                        break;
                    default: {

                    }

                }
            }
        });
    }

    @Deprecated
    public DictObjField getField(int index) {
        if (index >= this.dictObj.getColumns().size() || index < 0) {
            throw new DustDictException("列不存在");
        }
        return new DictObjField(this, this.dictObj.getColumn(index));
    }

    /**
     * 获取单元信息
     * @deprecated 请使用 getValue和setValue进行操作 
     * @see DictObjRow#setValue(String, Object)
     * @param name
     * @return
     */
    @Deprecated
    public DictObjField getField(String name) {
        checkNotNull(name);
        DictObjColumn col = this.dictObj.getColumn(name);
        if (col == null) {
            throw new DustDictException(this.dictObj.getCode() + '.' + name + "列不存在");
        }

        return new DictObjField(this, col);
    }


    @Deprecated
    public DictObjField get(int index) {
        return getField(index);
    }

    public void set(int index, Object value) {
        DictObjField f = getField(index);
        f.setValue(value);
    }

    @Deprecated
    public DictObjField get(String columnName) {
        return getField(columnName);
    }

    public Object getValue(String columnName) {
        return getValue(columnName, false);
    }

    public Object getValue(String columnName, boolean original) {
        DictObjColumn col = dictObj.getColumn(columnName);
        return getValue(col, original);
    }

    public Object getValue(DictObjColumn col) {
        return getValue(col, false);
    }

    public Object getValue(DictObjColumn col, boolean original) {
        if (col == null) {
            throw new DustDictException("列无效");
        }
        return col.toValue(data.getValue(col.getColumnName(), original));
    }

    public void setValue(String columnName, Object value) {
        DictObjColumn col = dictObj.getColumn(columnName);
        if (col == null) {
            throw new DustDictException("无效的列名称:" + columnName);
        }
        this.setValue(col, value, null);
    }

    public void setValue(DictObjColumn column, Object value, Object refValue) {
        if (this.data == null) {
            this.data = this.dictObj.dataTable.addRow();
        }
        Object colVal = column.toValue(value);
        Object originValue = this.getValue(column, true);

        boolean hasModify = (colVal == null && originValue != null)
                || (colVal != null && !colVal.equals(originValue));
        if (hasModify) {
            if (this.rowState == RowState.UNCHANGED) {
                this.setState(RowState.MODIFIED);
            }
        }

        Object oldValue = this.getValue(column);
        this.data.setValue(column.getColumnName(), colVal);
        if (refValue != null) {
            this.data.setValue(column.getColumnName(), refValue);
        }

        boolean changed = (oldValue == null && colVal != null)
                || (oldValue != null && !oldValue.equals(colVal));
        if (changed) {
            List<String> relatedFields = this.dictObj.getColumns().getRelationColumns(column.getColumnName());
            if (relatedFields != null) {
                relatedFields.forEach(f -> setValue(f, colVal));
            }

        }

        if (StringUtils.equals(column.getColumnName().toUpperCase(), DictConstant.C_ID.toUpperCase())) {
            this.setId(this.getLongValue(column.getColumnName()));
        }
        getDictObj().onValueChanged(this, column, this.getValue(column), oldValue);
    }

    public String getText(String name) {
        DictObjColumn col = this.dictObj.getColumn(name);
        if (col == null) {
            return StringUtils.EMPTY;
        }

        String refValue = getRefValue(col);
        if (StringUtils.isEmpty(refValue)) {
            return getStringValue(name);
        }

        return getStringValue(name) + DictConstant.REF_SEPARATOR + refValue;
    }

    private String getRefValue(DictObjColumn column) {
        return TypeUtil.toString(this.data.getValue(column.getRefNameColumn()));
    }

    public String getStringValue(String name) {
        return TypeUtil.toString(getValue(name));
    }

    public boolean getBooleanValue(String name) {
        return TypeUtil.toBoolean(getValue(name));
    }

    public int getInt32Value(String name) {
        return TypeUtil.toInteger(getValue(name));
    }

    public long getLongValue(String name) {
        return TypeUtil.toLong(getValue(name));
    }

    public Date getDateTimeValue(String name) {
        return TypeUtil.toDateTime(getValue(name));
    }

    public double getDoubleValue(String name) {
        return TypeUtil.toDouble(getValue(name));
    }

    public BigDecimal getDecimalValue(String name) {
        return TypeUtil.toDecimal(getValue(name));
    }

    public void setValue(DictObjColumn column, Object value) {
        setValue(column, value, null);
    }

    public void remove() {
        this.dictObj.removeRow(this);
    }

    public void delete() {
        dictObj.deleteRow(this);
    }

    public void rejectChanges() {
        this.data.rejectChanges();
        this.rowState = RowState.UNCHANGED;
    }

    public void acceptChanges() {
        this.data.acceptChanges();
        if (this.rowState == RowState.DETACHED || this.rowState == RowState.DELETED) {
            return;
        }
        this.rowState = RowState.UNCHANGED;
    }

    public List<DictObjField> getChanges() {
        List<DictObjField> arr = new ArrayList<>();
        for (DictObjColumn column : this.dictObj.getColumns()) {
            if (data.isChanged(column.getColumnName())) {
                arr.add(new DictObjField(this, column));
            }
        }

        return arr;
    }

    public void loadData(HDataRow row) {
        data = row;
        this.dictObj.getColumns().forEach(col -> {
            List<String> relCols = this.dictObj.getColumns().getRelationColumns(col.getColumnName());
            if (relCols != null) {
                relCols.forEach(relCol -> {
                    data.setValue(relCol, data.getValue(col.getColumnName()));
                });
            }
        });

        if (this.dictObj.getColumns().contains(DictConstant.C_ID)) {
            this.id = this.get(DictConstant.C_ID).getLongValue();
        }
    }

    public Map<String, Object> toMap() {
        Map<String, Object> map = new HashMap<>();
        for (DictObjColumn col : this.dictObj.getColumns()) {
            map.put(col.getColumnName(), getValue(col));
            map.put(col.getRefColumnName(), getRefValue(col));
        }
        return map;
    }

    public JSONArray toJSONArray() {
        return this.data.toJSONArray();
    }

    public JSONObject toJSON() {
        JSONObject o = new JSONObject();
        for (DictObjColumn col : this.dictObj.getColumns()) {
            if (col.getDataType() == DataType.UBIGINT) {
                o.put(col.getColumnName(), getStringValue(col.getColumnName()));
            } else {
                o.put(col.getColumnName(), getValue(col.getColumnName()));
            }

            if (!StringUtils.isEmpty(getRefValue(col))) {
                o.put(col.getRefColumnName(), getRefValue(col));
            }
        }
        return o;
    }

    public <T> T to(Class<T> type) {
        return ClassUtil.mapToObject(toMap(), type);
    }


    @Override
    public Iterator<Object> iterator() {
        return data.iterator();
    }

    private boolean validate(DictObjColumn column) {
        if (column.isAutoIncrement())  { return true; }

        Object value = this.getValue(column);
        if (column.isRequired() || column.isPrimaryKey()) {
            if (value == null || StringUtils.EMPTY.equals(value)) {
                Error error = new Error();
                error.setColumnName(column.getColumnName());
                error.setMessage(column.getColumnLabel() + "字段必填");
                errors.add(error);
                return false;
            }
        }

        boolean needValidateWidth = column.getDataWidth() > 0
                && (column.getDataType() == DataType.STRING
                || column.getDataType() == DataType.TEXT);
        if (needValidateWidth) {
            String str = getStringValue(column.getColumnName());
            int len = StringUtils.isEmpty(str) ? 0 : str.length();
            if (len > column.getDataWidth()) {
                Error error = new Error();
                error.setColumnName(column.getColumnName());
                error.setMessage(column.getColumnLabel() + "字段内容超长");
                errors.add(error);
                return false;
            }
        }

        return true;
    }

    public boolean validate() {
        this.errors.clear();
        boolean valid = true;
        for (DictObjColumn column : this.dictObj.getColumns()) {
            valid = validate(column) && valid;
        }

        return valid;
    }

    public HDataRow getData() {
        return data;
    }

    public void merge(DictObjRow src) {
        this.data.merge(src.data);
    }


    @Override
    public void dispose() {
        dictObj = null;
    }

    @Override
    protected void finalize() throws Throwable {
        super.finalize();
        this.dispose();
    }
}
