package top.daxian1218.dust.db.dict;

import top.daxian1218.dust.commons.exception.DustSystemException;
import top.daxian1218.dust.commons.generic.IDisposable;
import top.daxian1218.dust.commons.generic.Preconditions;
import top.daxian1218.dust.db.dict.event.ActionType;

import java.util.*;

/**
 * @author hst
 */
public class DictObjRowCollection implements List<DictObjRow>, IDisposable {
    private final List<DictObjRow> data = new ArrayList<>();
    private DictObj dictObj;

    public DictObjRowCollection(DictObj obj) {
        dictObj = obj;
    }

    @Override
    public int size() {
        return data.size();
    }

    @Override
    public boolean isEmpty() {
        return data.isEmpty();
    }

    @Override
    public boolean contains(Object o) {
        return data.contains(o);
    }

    @Override
    public Iterator<DictObjRow> iterator() {
        return data.iterator();
    }

    @Override
    public Object[] toArray() {
        return data.toArray();
    }

    @Override
    public <T> T[] toArray(T[] a) {
        return data.toArray(a);
    }

    @Override
    public boolean add(DictObjRow dictObjRow) {
        return add(dictObjRow, RowState.ADDED);
    }

    public boolean add(DictObjRow dictObjRow, RowState rowState) {
        Preconditions.checkNotNull(dictObjRow);
        if (dictObjRow.getState() != RowState.DETACHED || dictObjRow.getDictObj() != dictObj) {
            throw new DustSystemException("dictObjRow参数不符合要求");
        }

        if (rowState != RowState.ADDED
                && rowState != RowState.MODIFIED
                && rowState != RowState.UNCHANGED) {
            throw new DustSystemException("集合只能添加修改、新增、无变化行");
        }

        if (data.contains(dictObjRow)) {
            return false;
        }

        data.add(dictObjRow);
        dictObjRow.setState(rowState);
        dictObj.onCollectionChanged(ActionType.ADD);
        return true;
    }

    public void load(DictObjRow dictObjRow) {
        Preconditions.checkNotNull(dictObjRow);
        if (dictObjRow.getState() != RowState.DETACHED || dictObjRow.getDictObj() != dictObj) {
            throw new DustSystemException("dictObjRow参数不符合要求");
        }

        if (data.contains(dictObjRow)) {
            throw new DustSystemException("行已存在");
        }

        data.add(dictObjRow);
        dictObjRow.acceptChanges();
        dictObjRow.setState(RowState.UNCHANGED);
    }

    @Override
    public boolean remove(Object o) {
        Preconditions.checkNotNull(o);
        if (!(o instanceof DictObjRow)) {
            throw new DustSystemException("移除行必须是DictObjRow");
        }

        DictObjRow dictObjRow = (DictObjRow) o;
        if (!data.contains(dictObjRow)) {
            return false;
        }

        data.remove(dictObjRow);
        dictObjRow.setState(RowState.DETACHED);
        dictObj.onRowRemoved(dictObjRow);
        dictObj.onCollectionChanged(ActionType.REMOVE);
        return true;
    }

    @Override
    public boolean containsAll(Collection<?> c) {
        return data.containsAll(c);
    }

    @Override
    public boolean addAll(Collection<? extends DictObjRow> c) {
        Preconditions.checkNotNull(c);

        for (DictObjRow row : c) {
            if (row.getState() != RowState.DETACHED || row.getDictObj() != dictObj) {
                throw new DustSystemException("新增行不符合要求");
            }

            if (data.contains(row)) {
                throw new DustSystemException("行已存在");
            }

            row.setState(RowState.ADDED);
        }

        data.addAll(c);
        dictObj.onCollectionChanged(ActionType.ADD);
        return true;
    }

    @Override
    public boolean addAll(int index, Collection<? extends DictObjRow> c) {
        Preconditions.checkNotNull(c);

        for (DictObjRow row : c) {
            if (row.getState() != RowState.DETACHED || row.getDictObj() != dictObj) {
                throw new DustSystemException("新增行不符合要求");
            }

            if (data.contains(row)) {
                throw new DustSystemException("行已存在");
            }

            row.setState(RowState.ADDED);
        }

        data.addAll(index, c);
        dictObj.onCollectionChanged(ActionType.ADD);
        return true;
    }

    @Override
    public boolean removeAll(Collection<?> c) {
        for (Object o : c) {
            remove(o);
        }
        return true;
    }

    @Override
    public boolean retainAll(Collection<?> c) {
        int len = data.size();
        for (int i = len - 1; i >= 0; i--) {
            DictObjRow r = data.get(i);
            if (c.contains(r)) {
                continue;
            }

            remove(r);
        }
        return true;
    }

    @Override
    public void clear() {
        data.clear();
        dictObj.onCollectionChanged(ActionType.RESET);
    }

    @Override
    public DictObjRow get(int index) {
        if (index >= 0 && index < data.size()) {
            return data.get(index);
        }
        return null;
    }

    @Override
    public DictObjRow set(int index, DictObjRow row) {

        Preconditions.checkNotNull(row);
        if (row.getState() != RowState.DETACHED || row.getDictObj() != dictObj) {
            throw new DustSystemException("新增行不符合要求");
        }

        if (data.contains(row)) {
            throw new DustSystemException("行已存在");
        }

        if (index >= 0 && index < data.size()) {
            data.set(index, row);
        }
        throw new DustSystemException("index超出集合边界");
    }

    @Override
    public void add(int index, DictObjRow dictObjRow) {
        Preconditions.checkNotNull(dictObjRow);
        if (dictObjRow.getState() != RowState.DETACHED || dictObjRow.getDictObj() != dictObj) {
            throw new DustSystemException("dictObjRow参数不符合要求");
        }

        if (data.contains(dictObjRow)) {
            throw new DustSystemException("行已存在");
        }

        data.add(index, dictObjRow);
        dictObjRow.setState(RowState.ADDED);
        dictObj.onCollectionChanged(ActionType.ADD);
    }

    @Override
    public DictObjRow remove(int index) {
        DictObjRow row = data.get(index);
        remove(row);
        return row;
    }

    @Override
    public int indexOf(Object o) {
        return data.indexOf(o);
    }

    @Override
    public int lastIndexOf(Object o) {
        return data.lastIndexOf(o);
    }

    @Override
    public ListIterator<DictObjRow> listIterator() {
        return data.listIterator();
    }

    @Override
    public ListIterator<DictObjRow> listIterator(int index) {
        return data.listIterator(index);
    }

    @Override
    public List<DictObjRow> subList(int fromIndex, int toIndex) {
        return data.subList(fromIndex, toIndex);
    }

    public DictObj getDictObj() {
        return dictObj;
    }

    @Override
    public void dispose() {
        data.clear();
        dictObj = null;
    }
}
