package top.daxian1218.dust.db.dict;

import org.apache.commons.lang3.StringUtils;
import top.daxian1218.dust.commons.data.HDataColumn;
import top.daxian1218.dust.commons.data.StatType;
import top.daxian1218.dust.commons.generic.Preconditions;
import top.daxian1218.dust.commons.util.TypeUtil;
import top.daxian1218.dust.db.dict.condition.DictCondition;
import top.daxian1218.dust.db.dict.condition.DictExpression;
import top.daxian1218.dust.db.dict.condition.expressions.*;
import top.daxian1218.dust.db.sql.HSqlCommand;

import java.util.List;

import static top.daxian1218.dust.db.sql.SqlConstants.*;

public class SelectSqlGenerator {

    private final DictObj dictObj;

    public SelectSqlGenerator(DictObj obj) {
        Preconditions.checkNotNull(obj);
        dictObj = obj;
    }

    /**
     * 获取所有的查询条件
     * @return sql where条件
     */
    public HSqlCommand getWhere() {
        return getWhere(null);
    }

    /**
     * 获取所有的查询条件
     * @param appendCdt 追加条件
     * @return sql where条件
     */
    public HSqlCommand getWhere(DictCondition appendCdt) {
        HSqlCommand whereCmd = new HSqlCommand();
        if (dictObj.getConditions().size() > 0) {
            HSqlCommand cmd = getConditionsSql(dictObj.getConditions());
            whereCmd.appendWhere(LeftParen + cmd.getWhere() + RightParen);
            whereCmd.appendParameters(cmd.getParameters());
        }

        if (dictObj.getFixCondition() != null) {
            HSqlCommand cmd = getConditionSql(dictObj.getFixCondition());
            whereCmd.appendWhere(LeftParen + cmd.getWhere() + RightParen);
            whereCmd.appendParameters(cmd.getParameters());
        }

        if (!StringUtils.isEmpty(dictObj.getConditionSql())) {
            whereCmd.appendWhere(LeftParen + dictObj.getConditionSql() + RightParen);
        }

        if (!StringUtils.isEmpty(dictObj.getFixConditionSql())) {
            whereCmd.appendWhere(LeftParen + dictObj.getFixConditionSql() + RightParen);
        }

        if (appendCdt != null) {
            HSqlCommand cmd = getConditionSql(appendCdt);
            whereCmd.appendWhere(LeftParen + cmd.getWhere() + RightParen);
            whereCmd.appendParameters(cmd.getParameters());
        }

        return whereCmd;
    }

    public HSqlCommand getSubSql() {
        HSqlCommand selectCmd = new HSqlCommand(Select);
        selectCmd.appendSql(Blank);
        selectCmd.appendSql(ConstColumn);
        selectCmd.appendSql(Blank);
        selectCmd.appendSql(From)
                .appendSql(dictObj.getTableName())
                .appendSql(dictObj.getCode());
        joinTable(selectCmd, dictObj.getRelationTables());
        HSqlCommand whereCmd = getWhere();
        if (whereCmd.hasWhere()) {
            selectCmd.append(whereCmd);
        }
        return selectCmd;
    }


    public HSqlCommand generate() {
        return generate(-1);
    }

    public HSqlCommand generate(int pollyIndex) {
        HSqlCommand selectCmd = new HSqlCommand(Select);
        selectCmd.appendSql(Blank);
        if (dictObj.getColumns().size() == 0) {
            throw new DustDictException("字典列未定义");
        }

        boolean needComma = false;
        selectCmd.getFields().clear();
        for (DictObjColumn col : dictObj.getColumns()) {
            if (col.isIgnore() || StringUtils.isEmpty(col.getPhysicalName())) {
                continue;
            }

            if (needComma) {
                selectCmd.appendSql(Comma);
            }

            selectCmd.appendSql(col.getTableName(), false)
                    .appendSql(Dot, false)
                    .appendSql(col.getPhysicalName(), false)
                    .appendSql(As)
                    .appendSql(col.getColumnName());

            if (!StringUtils.isEmpty(col.getRelationColumn())) {
                String table = StringUtils.isEmpty(col.getRelationTable()) ? col.getTableName() : col.getRelationTable();
                selectCmd.appendSql(Comma);
                selectCmd.appendSql(table, false)
                        .appendSql(Dot, false)
                        .appendSql(col.getRelationColumn(), false)
                        .appendSql(As)
                        .appendSql(col.getColumnName() + "$");
            }

            if (col.getStatType() != StatType.NONE) {
                HDataColumn totalCol = new HDataColumn(col.getColumnName());
                totalCol.setStatType(col.getStatType());
                selectCmd.getFields().add(totalCol);
            }

            needComma = true;
        }

        selectCmd.appendSql(Blank);
        selectCmd.appendSql(From)
                .appendSql(dictObj.getTableName())
                .appendSql(dictObj.getCode());
        joinTable(selectCmd, dictObj.getRelationTables());

        DictCondition pollyCdt = null;
        if (pollyIndex >= 0) {
            if (dictObj.getPollyConditions().size() <= pollyIndex) {
                throw new DustDictException("无法定位故障查询条件，请检查字典查询条件");
            }

            pollyCdt = dictObj.getPollyConditions().get(pollyIndex);
        }

        HSqlCommand whereCmd = getWhere(pollyCdt);
        if (whereCmd.hasWhere()) {
            selectCmd.append(whereCmd);
        }

        List<OrderBy> orders = dictObj.getOrderBys();
        if (orders.size() > 0) {
            StringBuilder orderSql = new StringBuilder();
            for (OrderBy orderBy : orders) {
                if (orderSql.length() > 0) {
                    orderSql.append(Comma);
                }

                orderSql.append(orderBy.getColumnName());
                orderSql.append(Blank);
                orderSql.append(orderBy.getOrderByType() == OrderByType.DESC ? Desc : Asc);
            }

            if (orderSql.length() > 0) {
                selectCmd.appendOrderString(orderSql.toString());
            }
        }

        if (!StringUtils.isEmpty(dictObj.getOrderBySql())) {
            selectCmd.appendOrderString(dictObj.getOrderBySql());
        }

        if (dictObj.getPageInfo() != null) {
            selectCmd.setPageIndex(dictObj.getPageInfo().getStart());
            selectCmd.setPageSize(dictObj.getPageInfo().getPageSize());
            selectCmd.setTotalRows(dictObj.getPageInfo().getTotalRows());
        }

        return selectCmd;
    }

    protected void joinTable(HSqlCommand selectCmd, List<DictObjTable> tables) {
        if (tables.size() == 0) {
            return;
        }

        for (DictObjTable tb : tables) {
            if (tb.isPrimary()) {
                continue;
            }

            switch (tb.getRelationType()) {
                case LEFT:
                    selectCmd.appendSql(LeftJoin);
                    break;
                case RIGHT:
                    selectCmd.appendSql(RightJoin);
                    break;
                case INNER:
                    selectCmd.appendSql(InnerJoin);
                    break;
                default:
                    selectCmd.appendSql(LeftJoin);
            }

            selectCmd.appendSql(tb.getTableName())
                    .appendSql(tb.getTableAlias());

            HSqlCommand onCmd = new HSqlCommand();
            List<DictCondition> conditions = tb.getConditions();
            if (conditions.size() > 0) {
                onCmd.append(getConditionsSql(conditions));
            }

            if (!StringUtils.isEmpty(tb.getThisColumn())
                    && !StringUtils.isEmpty(tb.getRelationColumn())) {
                StringBuilder onWhere = new StringBuilder();
                String relateTable = StringUtils.isEmpty(tb.getRelationTable()) ? dictObj.getCode()
                        : tb.getRelationTable();
                onWhere.append(tb.getTableAlias()).append(Dot).append(tb.getThisColumn())
                        .append("=").append(relateTable).append(Dot).append(tb.getRelationColumn());
                onCmd.appendWhere(onWhere.toString());
            }

            if (!StringUtils.isEmpty(tb.getRelationWhere())) {
                onCmd.appendWhere(tb.getRelationWhere());
            }

            if (onCmd.hasWhere()) {
                selectCmd.appendSql(On);
                selectCmd.appendSql(onCmd.getWhere());
                selectCmd.appendParameters(onCmd.getParameters());
            }

        }
    }

    private HSqlCommand getConditionsSql(List<DictCondition> conditions) {
        HSqlCommand subWhereCmd = new HSqlCommand();
        for (DictCondition item : conditions) {
            HSqlCommand cmd = getConditionSql(item);
            subWhereCmd.append(cmd, !item.isRequire());
        }

        return subWhereCmd;
    }

    private void amendCondition(DictCondition cdt) {
        if (dictObj == null)  { return; }

        // 修复列查询有关的内容，修正列相关的表名以及值类型
        if (cdt.getLeft() instanceof ColumnExpression) {
            amendColumnExpr((ColumnExpression) cdt.getLeft(), cdt.getRight());
        }

        if (cdt.getRight() instanceof ColumnExpression) {
            amendColumnExpr((ColumnExpression) cdt.getRight(), cdt.getLeft());
        }
    }

    private DictObjColumn findColumn(String tableName, String physicalName) {
        for (DictObjColumn col : dictObj.getColumns()) {
            if (StringUtils.equals(col.getPhysicalName(), physicalName) &&
                    StringUtils.equals(col.getTableName(), tableName)) {
                return col;
            }
        }

        return null;
    }

    private void amendColumnExpr(ColumnExpression ce, DictExpression expr) {

        // 如果表名为空，则需要检索对象列，找到符合要求的列。或者补全table name为对象表
        if (!StringUtils.isEmpty(ce.getTableName())) {
            // 隶属于字典表，则修正值类型
            if (StringUtils.equals(ce.getTableName(), dictObj.getCode())) {
                DictObjColumn matchColumn = findColumn(ce.getTableName(), ce.getColumnName());
                if (matchColumn != null) {
                    amendExpressionDataType(expr, matchColumn);
                }
            }
            return;
        }

        DictObjColumn column = dictObj.getColumns().get(ce.getColumnName());
        if (column != null) {
            ce.setColumnName(column.getPhysicalName());
            ce.setTableName(column.getTableName());
            amendExpressionDataType(expr, column);
        } else {
            ce.setTableName(dictObj.getCode());
        }
    }

    public void amendExpressionDataType(DictExpression expr, DictObjColumn column) {
        if (expr instanceof ValueExpression) {
            ValueExpression ve = (ValueExpression) expr;
            ve.setDataType(column.getDataType());
        }

        if (expr instanceof ValueListExpression) {
            ValueListExpression vle = (ValueListExpression) expr;
            vle.setDataType(column.getDataType());
        }
    }

    protected HSqlCommand getConditionSql(DictCondition condition) {
        HSqlCommand cmd = new HSqlCommand();
        amendCondition(condition);
        boolean skipCondition = condition.getLeft() == null && condition.getRight() == null;
        if (!skipCondition) {
            switch (condition.getOperation()) {
                case EQUAL:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql("=");
                    appendColumnOrValueSql(condition.getRight(), cmd);
                    break;
                case NOT_EQUAL:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql("<>");
                    appendColumnOrValueSql(condition.getRight(), cmd);
                    break;
                case GREATER:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql(">");
                    appendColumnOrValueSql(condition.getRight(), cmd);
                    break;
                case EQUAL_GREATER:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql(">=");
                    appendColumnOrValueSql(condition.getRight(), cmd);
                    break;
                case LESS:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql("<");
                    appendColumnOrValueSql(condition.getRight(), cmd);
                    break;
                case EQUAL_LESS:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql("<=");
                    appendColumnOrValueSql(condition.getRight(), cmd);
                    break;
                case NULL:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql(IsNull);
                    break;
                case NOT_NULL:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql(IsNotNull);
                    break;
                case CONTAIN:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql(Like);
                    appendFuzzyValueSql(condition.getRight(), cmd, 3);
                    break;
                case START_WITH:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql(Like);
                    appendFuzzyValueSql(condition.getRight(), cmd, 2);
                    break;
                case END_WITH:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql(Like);
                    appendFuzzyValueSql(condition.getRight(), cmd, 1);
                    break;
                case IN_LIST:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql(In);
                    appendValueListSql(condition.getRight(), cmd);
                    break;
                case NOT_IN:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql(Not);
                    cmd.appendSql(In);
                    appendValueListSql(condition.getRight(), cmd);
                    break;
                case IN_TABLE:
                    appendColumnOrValueSql(condition.getLeft(), cmd);
                    cmd.appendSql(In).appendSql("(");
                    appendSubTableSql(condition.getRight(), cmd);
                    cmd.appendSql(")");
                    break;
                case EXIST:
                    cmd.appendSql(Exists).appendSql("(");
                    appendSubTableSql(condition.getRight(), cmd);
                    cmd.appendSql(")");
                    break;
                case NOT_EXIST:
                    cmd.appendSql(Not).appendSql(Exists).appendSql("(");
                    appendSubTableSql(condition.getRight(), cmd);
                    cmd.appendSql(")");
                    break;
            }


        }

        if (condition.getSubConditions().size() > 0) {
            HSqlCommand otherCmd = getSubConditionsCommand(condition.getSubConditions());

            if (!skipCondition) {
                cmd.appendSql(And);
            }

            if (condition.getSubConditions().size() == 1) {
                cmd.appendSql(otherCmd.getWhere());
            }
            else {
                cmd.appendSql("(" + otherCmd.getWhere() + ")");
            }
            cmd.appendParameters(otherCmd.getParameters());
        }


        //转化commandText为where
        cmd.setWhere(StringUtils.EMPTY);
        cmd.appendWhere(cmd.getCommandText());
        cmd.resetSql();
        return cmd;
    }

    protected void appendSubTableSql(DictExpression expr, HSqlCommand cmd) {
        if (expr instanceof SubTableExpression) {
            SubTableExpression ste = (SubTableExpression) expr;
            cmd.appendSql(Select);
            cmd.appendSql(!StringUtils.isEmpty(ste.getColumnName()) ? ste.getAlias() + Dot + ste.getColumnName() :ConstColumn);
            cmd.appendSql(From);
            cmd.appendSql(ste.getTableName());
            cmd.appendSql(ste.getAlias());

            List<DictCondition> subConditions = ste.getSubConditions();
            if (subConditions.size() <= 0) {
                return;
            }

            cmd.appendSql(Where);
            HSqlCommand subCmd = getSubConditionsCommand(subConditions);
            cmd.appendSql(subCmd.getWhere(), true);
            cmd.appendParameters(subCmd.getParameters());
        }

        if (expr instanceof CommandExpression) {
            CommandExpression cer = (CommandExpression)expr;
            cmd.appendSql(cer.getCommand().getCommandText());
            cmd.appendSql(Where);
            cmd.appendSql(cer.getCommand().getWhere(), true);
            cmd.appendParameters(cer.getCommand().getParameters());
        }

        if (expr instanceof SqlExpression) {
            SqlExpression se = (SqlExpression)expr;
            cmd.appendSql(se.getSql());
        }
    }

    protected void appendColumnOrValueSql(DictExpression expr, HSqlCommand cmd) {
        switch (expr.getClass().getSimpleName()) {
            case "ColumnExpression":
                ColumnExpression ce = (ColumnExpression) expr;
                cmd.appendSql(ce.getTableName() + Dot + ce.getColumnName());
                break;
            case "ValueExpression":
                ValueExpression ve = (ValueExpression) expr;
                cmd.appendSql("${INDEX}");
                cmd.appendParameter(ve.getValue());
                break;
        }
    }

    /**
     * 模糊匹配，mode=2 首匹配 mode=1 末尾匹配 mode=3 包含匹配
     *
     * @param expr
     * @param cmd
     * @param mode
     */
    protected void appendFuzzyValueSql(DictExpression expr, HSqlCommand cmd, int mode) {
        StringBuilder sbValue = new StringBuilder();
        if (expr instanceof ValueExpression) {
            ValueExpression ve = (ValueExpression) expr;
            if ((mode & 0x1) == 1) {
                sbValue.append("%");
            }

            sbValue.append(TypeUtil.toString(ve.getValue()));

            cmd.appendSql("${INDEX}");
            if ((mode & 0x2) == 2) {
                sbValue.append("%");
            }
            cmd.appendParameter(sbValue.toString());
        }
    }


    protected void appendValueListSql(DictExpression node, HSqlCommand cmd) {
        cmd.appendSql("(");
        if (node instanceof ValueListExpression)
        {
            ValueListExpression expr = (ValueListExpression) node;
            Object[] values = expr.getValues();
            for (int i = 0; i < values.length; i++) {
                if (i > 0) {
                    cmd.appendSql(Comma);
                }
                cmd.appendSql("${INDEX}");
                cmd.appendParameter(values[i]);
            }
        }

        cmd.appendSql(")");
    }

    private HSqlCommand getSubConditionsCommand(List<DictCondition> subConditions) {
        HSqlCommand cmd = new HSqlCommand();
        if (subConditions.size() == 0) {
            return cmd;
        }

        for(DictCondition condition : subConditions)
        {
            HSqlCommand subCmd = getConditionSql(condition);
            cmd.append(subCmd, !condition.isRequire());
        }

        return cmd;
    }

}
