package top.daxian1218.dust.db.dict;

import org.apache.commons.lang3.StringUtils;
import top.daxian1218.dust.commons.data.HDataColumn;
import top.daxian1218.dust.commons.data.HDataTable;
import top.daxian1218.dust.commons.generic.Preconditions;
import top.daxian1218.dust.db.DictGlobalConfig;
import top.daxian1218.dust.db.base.IDataAdapter;
import top.daxian1218.dust.db.sql.HSqlCommand;
import top.daxian1218.dust.db.sql.command.HDeleteSqlCommand;
import top.daxian1218.dust.db.sql.command.HInsertSqlCommand;
import top.daxian1218.dust.db.sql.command.HUpdateSqlCommand;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;


/**
 * @author hst
 */
public abstract class SqlBuilder {

    public abstract boolean isSupport(String dbType);
    public String getSelectWhere(DictObj obj) {
        String where = new SelectSqlGenerator(obj).getWhere().getDbExecuteSql((v, i) -> getDbValue(v));
        if (where.length() > 7) {
            where = where.substring(7);
        }
        return where;
    }

    protected abstract String getDbValue(Object o);

    protected SelectSqlGenerator getSelectGenerator(DictObj obj) {
        return new SelectSqlGenerator(obj);
    }

    public String getSelectSql(DictObj obj) {
        HSqlCommand selectCmd = getSelectGenerator(obj).generate();
        return selectCmd.getDbExecuteSql((v, i) -> getDbValue(v));
    }

    public String getInsertSql(DictObj destObj, String tbName) {
        DictObjTable tb = destObj.getTable(tbName);
        HSqlCommand insertCmd = getInsertSqlCore(destObj, tb);
        return insertCmd.getDbExecuteSql((v, i) -> getDbValue(v));
    }

    public String getUpdateSql(DictObj destObj, String tbName) {
        DictObjTable tb = destObj.getTable(tbName);
        HSqlCommand updateCmd = getUpdateSqlCore(destObj, tb);
        return updateCmd.getDbExecuteSql((v, i) -> getDbValue(v));
    }

    /**
     * 自定义处理插入列
     *
     * @param col
     * @param cmd
     * @return
     */
    protected boolean handleInsertSqlCoreColumn(DictObjColumn col, HInsertSqlCommand cmd) {
        return col.isAutoIncrement() || col.isIgnore() || StringUtils.isEmpty(col.getPhysicalName()) ||
                StringUtils.equals(col.getDefaultValue(), DictConstant.DEFAULT_VALUE_NOW) ||
                StringUtils.equals(col.getDefaultValue(), DictConstant.DEFAULT_VALUE_UPDATE);
    }

    /**
     * 自定义处理更新列
     *
     * @param col
     * @param cmd
     * @return
     */
    protected boolean handleUpdateSqlCoreColumn(DictObjColumn col, HUpdateSqlCommand cmd) {
        return col.isIgnore() || StringUtils.isEmpty(col.getPhysicalName())
                || StringUtils.equals(col.getDefaultValue(), DictConstant.DEFAULT_VALUE_NOW)
                || col.isAutoIncrement();
    }


    protected HSqlCommand getInsertSqlCore(DictObj destObj, DictObjTable tb) {
        HInsertSqlCommand cmd = new HInsertSqlCommand(tb.getTableName());
        destObj.getColumns().forEach(col ->
        {
            if (StringUtils.equals(col.getTableName(), tb.getTableAlias())) {
                if (!handleInsertSqlCoreColumn(col, cmd)) {
                    HDataColumn dc = new HDataColumn(col.getPhysicalName());
                    dc.setAlias(col.getColumnName());
                    cmd.getFields().add(dc);
                }
            }
        });

        return cmd;
    }


    protected HSqlCommand getUpdateSqlCore(DictObj destObj, DictObjTable tb) {
        HUpdateSqlCommand cmd = new HUpdateSqlCommand(tb.getTableName());
        destObj.getColumns().forEach(col ->
        {
            if (StringUtils.equals(col.getTableName(), tb.getTableAlias())) {
                if (col.isPrimaryKey() || !handleUpdateSqlCoreColumn(col, cmd)) {
                    HDataColumn dc = new HDataColumn(col.getPhysicalName());
                    dc.setAlias(col.getColumnName());
                    dc.setPrimaryKey(col.isPrimaryKey());
                    cmd.getFields().add(dc);
                }
            }
        });

        return cmd;
    }

    protected HSqlCommand getDeleteSqlCore(DictObj destObj, DictObjTable tb) {
        HDeleteSqlCommand cmd = new HDeleteSqlCommand(tb.getTableName());
        destObj.getColumns().forEach(col ->
        {
            if (!col.isIgnore() && col.isPrimaryKey()
                    && StringUtils.equals(col.getTableName(), tb.getTableAlias())) {
                cmd.appendWhere(col.getPhysicalName() + "=:" + col.getPhysicalName());
                HDataColumn dc = new HDataColumn(col.getPhysicalName());
                dc.setAlias(col.getColumnName());
                dc.setPrimaryKey(col.isPrimaryKey());
                cmd.getFields().add(dc);
            }
        });

        return cmd.getFields().size() > 0 ? cmd : null;
    }

    public StringBuilder BatchInsertSql(DictObj destObj, DictObjTable tb) {
        HSqlCommand insertCmd = getInsertSqlCore(destObj, tb);
        StringBuilder batchSql = new StringBuilder();

        insertCmd.iterate(cmd ->
        {
            batchSql.append(insertCmd.getDbExecuteSql((v, i) -> getDbValue(v)));
            batchSql.append("\n");
            return true;
        });
        return batchSql;
    }

    public StringBuilder batchUpdateSql(DictObj destObj, String tbName) {
        DictObjTable tb = destObj.getTable(tbName);
        HSqlCommand updateCmd = getUpdateSqlCore(destObj, tb);
        StringBuilder batchSql = new StringBuilder();
        updateCmd.iterate(cmd ->
        {
            batchSql.append(updateCmd.getDbExecuteSql((v, i) -> getDbValue(v)));
            batchSql.append("\n");
            return true;
        });
        return batchSql;
    }

    /**
     * 根据SqlCommand的列信息将DictObjRow转化为参数
     *
     * @param row
     * @param cmd
     */
    protected void filterCommandParameter(DictObjRow row, HSqlCommand cmd) {
        if (cmd == null) {
            return;
        }

        // 新增一个执行命令
        cmd.applyBatch();
        for (HDataColumn c : cmd.getFields()) {
            if (c.isIgnore()) {
                continue;
            }

            if (c.isPrimaryKey()) {
                cmd.setParameter(HSqlCommand.PrefixOld + c.getName(), row.getValue(c.getAlias(), true));
            }

            cmd.setParameter(c.getName(), row.getValue(c.getAlias()));
        }
    }

    /**
     * 执行SqlCommand，如果SqlCommand没有参数，则不需要执行;用于insert和update
     *
     * @param cmd
     * @param adapter
     */
    protected void executeCommand(HSqlCommand cmd, IDataAdapter adapter) {
        if (cmd.hasParameter()) {
            adapter.update(cmd);
        }
    }

    public void search(DictObj destObj, IDataAdapter adapter) {
        Preconditions.checkNotNull(destObj);
        if (adapter == null) {
            adapter = DictGlobalConfig.Instance.getDataAdapter();
        }

        if (adapter == null) {
            throw new DustDictException("current thread can not find dict sql adapter");
        }

        // 支持Polly故障逻辑
        if (destObj.getPollyConditions().size() > 0) {
            if (destObj.getPollyType() == 0) {
                retrySearch(destObj, adapter);
                return;
            }
        }

        HSqlCommand selectCmd = getSelectGenerator(destObj).generate();

        HDataTable hDataTable = adapter.query(selectCmd);
        destObj.loadData(hDataTable);
        destObj.getPageInfo().setTotalRows(selectCmd.getTotalRows());
        selectCmd.getFields().forEach(c -> {
            destObj.getColumn(c.getName()).setStatValue(c.getStatValue());
        });
    }

    private void retrySearch(DictObj destObj, IDataAdapter adapter) {
        for(int i = 0; i < destObj.getPollyConditions().size(); i++) {
            HSqlCommand selectCmd = getSelectGenerator(destObj).generate(i);

            HDataTable hDataTable = adapter.query(selectCmd);
            if (hDataTable.size() == 0) {
                continue;
            }

            destObj.loadData(hDataTable);
            destObj.getPageInfo().setTotalRows(selectCmd.getTotalRows());
            selectCmd.getFields().forEach(c -> {
                destObj.getColumn(c.getName()).setStatValue(c.getStatValue());
            });
            return;
        }
    }

    public void save(DictObj destObj, IDataAdapter adapter, boolean autoCommit) {
        if (adapter == null) {
            adapter = DictGlobalConfig.Instance.getDataAdapter();
        }

        if (adapter == null) {
            throw new DustDictException("current thread can not find dict sql adapter");
        }

        List<HSqlCommand> deleteCommands = new ArrayList<>();
        List<HSqlCommand> insertCommands = new ArrayList<>();
        List<HSqlCommand> updateCommands = new ArrayList<>();
        deleteCommands.add(0, getDeleteSqlCore(destObj, destObj.getTable()));
        insertCommands.add(0, getInsertSqlCore(destObj, destObj.getTable()));
        updateCommands.add(0, getUpdateSqlCore(destObj, destObj.getTable()));

        for (DictObjTable tb : destObj.getRelationTables()) {
            if (tb.isFollowDelete()) {
                deleteCommands.add(getDeleteSqlCore(destObj, tb));
            }

            if (tb.isFollowInsert()) {
                insertCommands.add(getInsertSqlCore(destObj, tb));
            }

            if (tb.isFollowUpdate()) {
                updateCommands.add(getUpdateSqlCore(destObj, tb));
            }
        }

        // 生成参数
        destObj.forEachChanges(row ->
        {
            switch (row.getState()) {
                case DELETED: {
                    for (HSqlCommand t : deleteCommands) {
                        filterCommandParameter(row, t);
                    }

                    break;
                }
                case MODIFIED: {
                    for (HSqlCommand t : updateCommands) {
                        filterCommandParameter(row, t);
                    }

                    break;
                }
                case ADDED: {
                    for (HSqlCommand t : insertCommands) {
                        filterCommandParameter(row, t);
                    }

                    break;
                }
            }
        });

        // Step1: 执行删除操作
        for (HSqlCommand cmd : deleteCommands) {
            if (cmd == null) {
                continue;
            }

            executeCommand(cmd, adapter);
        }

        // Step2: 更新操作
        for (HSqlCommand cmd : updateCommands) {
            if (cmd == null) {
                continue;
            }

            executeCommand(cmd, adapter);
        }

        // Step3: 插入主体表， 获取自增id，更新镜像列
        HSqlCommand masterInsert = insertCommands.get(0);
        if (masterInsert.hasParameter()) {
            executeCommand(masterInsert, adapter);
            updateAutoIncrementColumn(destObj, adapter);
        }

        // Step4: 插入关联表数据
        for (int i = 1; i < insertCommands.size(); i++) {
            executeCommand(insertCommands.get(i), adapter);
        }

        if (autoCommit) {
            adapter.commit();
        }
    }

    private void updateAutoIncrementColumn(DictObj destObj, IDataAdapter adapter) {
        // 找出自增列
        String incrementColumn = StringUtils.EMPTY;
        for (DictObjColumn col : destObj.getColumns()) {
            if (col.isAutoIncrement() && StringUtils.equals(col.getTableName(), destObj.getTableName())) {
                incrementColumn = col.getColumnName();
                break;
            }
        }

        if (StringUtils.isEmpty(incrementColumn)) {
            return;
        }

        List<String> updatingColumns = Arrays.asList(incrementColumn);

        for (DictObjColumn col : destObj.getColumns()) {
            if (StringUtils.equals(col.getMirrorColumn(), incrementColumn)) {
                updatingColumns.add(col.getColumnName());
            }
        }

        long id = getLastId(adapter, destObj, incrementColumn);
        if (id <= 0) {
            return;
        }

        for (int i = 0; i < destObj.getRows().size(); i++) {
            for (String updatingColumn : updatingColumns) {
                destObj.getRows().get(i).get(updatingColumn).setValue(id - (destObj.getRows().size() - 1 - i));
            }
        }
    }

    protected abstract long getLastId(IDataAdapter adapter, DictObj destObj, String incrementColumn);
}
