package top.daxian1218.dust.db.dict.condition;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import org.apache.commons.lang3.StringUtils;
import top.daxian1218.dust.commons.util.TypeUtil;
import top.daxian1218.dust.db.dict.DictObj;
import top.daxian1218.dust.db.dict.DictObjColumn;
import top.daxian1218.dust.db.dict.DustDictException;
import top.daxian1218.dust.db.dict.condition.expressions.*;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * @author hst
 */
public class DictCondition {
    private boolean require = true;
    private List<DictCondition> subConditions = new ArrayList<>();
    private DictExpression left;
    private DictExpression right;
    private OperationType operation = OperationType.EQUAL;

    public DictCondition() {

    }

    public DictCondition(DictExpression left, OperationType operation, DictExpression right, boolean require) {
        this.left = left;
        this.operation = operation;
        this.right = right;
        this.require = require;
    }

    public DictCondition(String tableAlias, String columnName, Object value) {
        this.left = new ColumnExpression(columnName, tableAlias);
        this.operation = OperationType.EQUAL;
        this.right = new ValueExpression(value);
    }

    public DictCondition(String columnName, Object value) {
        this.left = new ColumnExpression(columnName);
        this.operation = OperationType.EQUAL;
        this.right = new ValueExpression(value);
    }

    public boolean isRequire() {
        return require;
    }

    public boolean isEmpty() { return left == null && right == null && subConditions.size() == 0; }

    public DictCondition setRequire(boolean require) {
        this.require = require;
        return this;
    }

    public List<DictCondition> getSubConditions() {
        return subConditions;
    }

    public DictExpression getLeft() {
        return left;
    }

    public DictCondition setLeft(DictExpression left) {
        this.left = left;
        return this;
    }

    public DictExpression getRight() {
        return right;
    }

    public DictCondition setRight(DictExpression right) {
        this.right = right;
        return this;
    }

    public OperationType getOperation() {
        return operation;
    }

    public DictCondition setOperation(OperationType operation) {
        this.operation = operation;
        return this;
    }

    private static DictExpression parseExpression(JSONObject nodeJson) {
        switch (nodeJson.getString("type")) {
            case "ColumnExpression":
                return new ColumnExpression().parse(nodeJson);
            case "SqlExpression":
                return new SqlExpression().parse(nodeJson);
            case "SubTableExpression":
                return new SubTableExpression().parse(nodeJson);
            case "ValueExpression":
                return new ValueExpression().parse(nodeJson);
            case "ValueListExpression":
                return new ValueListExpression().parse(nodeJson);
        }
        throw new DustDictException("无效类型");
    }

    public static DictCondition parse(JSONObject jsonObject) {
        DictCondition cdt = new DictCondition();
        String strOperation = jsonObject.getString("operation");
        if (strOperation == null) {
            throw new DustDictException("parsing JSONObject have no compatible operation");
        }

        cdt.operation = TypeUtil.toEnumByName(OperationType.class, strOperation);
        if (jsonObject.containsKey("require")) {
            cdt.require = jsonObject.getBoolean("require");
        } else if (jsonObject.containsKey("isRequire")) {
            cdt.require = jsonObject.getBoolean("isRequire");
        }

        JSONObject rightJson = jsonObject.getJSONObject("right");
        if (rightJson != null) {
            cdt.right = parseExpression(rightJson);
        }

        JSONObject leftJson = jsonObject.getJSONObject("left");
        if (leftJson != null) {
            cdt.left = parseExpression(leftJson);
        }

        //sub conditions
        JSONArray jaSub = jsonObject.getJSONArray("subConditions");
        if (jaSub != null) {
            for (int i = 0; i < jaSub.size(); i++) {
                JSONObject jo = jaSub.getJSONObject(i);
                if (jo != null) {
                    cdt.subConditions.add(parse(jo));
                }
            }
        }
        return cdt;
    }

    public JSONObject toJson() {
        JSONObject json = new JSONObject();
        json.put("operation", operation.toString());
        json.put("require", require);
        if (left != null) {
            json.put("left", left.toJson());
        }

        if (right != null) {
            json.put("right", right.toJson());
        }

        if (subConditions.size() > 0) {
            JSONArray subConditions = new JSONArray();
            subConditions.forEach(c -> {
                DictCondition dc = (DictCondition) c;
                subConditions.add(dc.toJson());
            });
            json.put("subConditions", subConditions);
        }

        return json;
    }

    public JSONObject toFieldValues() {
        JSONObject cur = toFieldValue();
        if (!subConditions.isEmpty()) {
            for (DictCondition sub : subConditions) {
                mergeFieldValue(cur, sub.toFieldValues());
            }
        }

        return cur;
    }

    private void mergeFieldValue(JSONObject des, JSONObject src) {
        for(String k : src.keySet()) {
            if (!des.containsKey(k)) {
                des.put(k, src.get(k));
                continue;
            }

            if (!(des.get(k) instanceof JSONArray)) {
                des.put(k, new JSONArray(Arrays.asList(des.get(k), src.get(k))));
                continue;
            }

            des.getJSONArray(k).add(src.get(k));
        }
    }

    private  JSONObject toFieldValue() {
        JSONObject json = new JSONObject();
        if (left instanceof ColumnExpression) {
            ColumnExpression colExpr = (ColumnExpression) left;
            String field = StringUtils.EMPTY;
            String value = StringUtils.EMPTY;

            switch (operation) {
                case GREATER:
                    field = ")" + colExpr.getColumnName();
                    break;
                case LESS:
                    field = "(" + colExpr.getColumnName();
                    break;
                case EQUAL_GREATER:
                    field = "]" + colExpr.getColumnName();
                    break;
                case EQUAL_LESS:
                    field = "[" + colExpr.getColumnName();
                    break;
                case NULL:
                    field = "#" + colExpr.getColumnName();
                    value = "false";
                    break;
                case NOT_NULL:
                    field = "#" + colExpr.getColumnName();
                    value = "true";
                    break;
                case EQUAL:
                    field = "=" + colExpr.getColumnName();
                    break;
                case NOT_EQUAL:
                    field = "!" + colExpr.getColumnName();
                    break;
                case CONTAIN:
                    field = "%" + colExpr.getColumnName();
                    break;
                case START_WITH:
                    field = "^" + colExpr.getColumnName();
                    break;
                case END_WITH:
                    field = "$" + colExpr.getColumnName();
                    break;
                case IN_LIST:
                    field = "@" + colExpr.getColumnName();
                    break;
                default:
                    break;
            }

            if (right instanceof ValueExpression) {
                value = TypeUtil.toString(((ValueExpression) right).getValue());
            }

            if (right instanceof ValueListExpression) {
                value = Arrays.stream(((ValueListExpression) right).getValues()).map(TypeUtil::toString).collect(Collectors.joining(","));
            }

            if (!StringUtils.isEmpty(field)) {
                json.put(field, value);
            }
        }

        return json;
    }

    public static DictCondition parseCondition(DictObj obj, String field, String value) {
        DictObjColumn col = obj.getColumn(field);
        if (col == null) {
            col = obj.getColumn(field.substring(1));
        }

        if (col == null) {
            return null;
        }

        if (col.isIgnore() || StringUtils.isEmpty(col.getTableName())) {
            return null;
        }

        DictCondition where = new DictCondition();
        where.setLeft(new ColumnExpression(col.getColumnName()));

        ValueExpression valueNode = new ValueExpression(value);
        valueNode.setDataType(col.getDataType());
        where.setRight(valueNode);

        char ch = field.charAt(0);
        switch (ch) {
            case ')':
                where.setOperation(OperationType.GREATER);
                break;
            case '(':
                where.setOperation(OperationType.LESS);
                break;
            case ']':
                where.setOperation(OperationType.EQUAL_GREATER);
                break;
            case '[':
                where.setOperation(OperationType.EQUAL_LESS);
                break;
            case '#':
                where.setOperation(TypeUtil.toBoolean(valueNode.getValue()) ?
                        OperationType.NOT_NULL : OperationType.NULL);
                break;
            case '=':
                where.setOperation(OperationType.EQUAL);
                break;
            case '!':
                where.setOperation(OperationType.NOT_EQUAL);
                break;
            case '%':
                where.setOperation(OperationType.CONTAIN);
                break;
            case '^':
                where.setOperation(OperationType.START_WITH);
                break;
            case '$':
                where.setOperation(OperationType.END_WITH);
                break;
            case '@':
                where.setOperation(OperationType.IN_LIST);
                ValueListExpression valueListNode = new ValueListExpression(Arrays.asList(value.split(",").clone()));
                valueListNode.setDataType(col.getDataType());
                where.setRight(valueListNode);
                break;
            default:
                where.setLeft(new ColumnExpression(col.getColumnName()));
                where.setOperation(OperationType.EQUAL);
                break;
        }
        return where;
    }
}
