package top.daxian1218.dust.db.sql;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import top.daxian1218.dust.commons.data.HDataTable;
import top.daxian1218.dust.commons.generic.Preconditions;
import top.daxian1218.dust.commons.log.LoggerHelper;
import top.daxian1218.dust.db.DustDbException;
import top.daxian1218.dust.db.base.*;
import top.daxian1218.dust.db.util.DbUtil;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.function.Consumer;

/**
 * 默认数据库适配器
 * 管理本次使用的数据源，连接
 *
 * @author hst
 */
public class DefaultDataAdapter implements IDataAdapter {
    private static Logger logger = LoggerFactory.getLogger(DefaultDataAdapter.class);

    private IDbOperations dbOperations;
    protected Connection connection;
    private IDataSource dataSource;
    private boolean isInitialized = false;
    private boolean abandon = true;

    /**
     * 初始化相应的组件并检查连接池是否存在
     *
     * @throws DustDbException
     */
    @Override
    public void init(IDataSource dataSource, IDbOperations dbOperations) {
        Preconditions.checkNotNull(dataSource);
        Preconditions.checkNotNull(dbOperations);

        if (isInitialized) {
            return;
        }
        this.dataSource = dataSource;
        this.dbOperations = dbOperations;
        isInitialized = true;
    }

    /**
     * 根据Command返回相应的查询结果
     *
     * @param cmd
     * @return {@link HDataTable}
     * @throws DustDbException
     */
    @Override
    public HDataTable query(HDataCommand cmd) throws DustDbException {
        HSqlCommand sqlCmd = (HSqlCommand) cmd;
        if (sqlCmd == null) {
            throw new DustDbException("非法的SQL命令");
        }
        open();
        sqlCmd.setConnection(this.connection);

        return dbOperations.query(sqlCmd, DbUtil::toTable);
    }

    @Override
    public void query(HDataCommand cmd, Consumer<ResultSet> handler) throws DustDbException {
        dbOperations.query(cmd, (rs) -> {
            handler.accept(rs);
            return null;
        });
    }

    /**
     * @param cmd
     * @return
     * @throws DustDbException
     */
    @Override
    public UpdateResult update(HDataCommand cmd) throws DustDbException {
        if (!(cmd instanceof HSqlCommand)) {
            throw new DustDbException("cmd must be sql command");
        }
        open();
        ((HSqlCommand) cmd).setConnection(this.connection);
        return dbOperations.update(cmd);
    }

    @Override
    public String getDbType() {
        return this.dataSource.getDbAccess().getDbType();
    }

    @Override
    public boolean isOpen() {
        try {
            return connection != null && !connection.isClosed();
        } catch (SQLException e) {
            return false;
        }
    }

    @Override
    public void open() throws DustDbException {
        if (isOpen()) {
            return;
        }

        try {
            this.connection = dataSource.getConnection();
            this.connection.setAutoCommit(false);
            if (dataSource.getDbAccess().getCluster() == DbCluster.ReadWrite) {
                this.connection.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
            } else if (dataSource.getDbAccess().getCluster() == DbCluster.ReadOnly) {
                this.connection.setTransactionIsolation(Connection.TRANSACTION_NONE);
            }
        } catch (SQLException e) {
            throw new DustDbException("connection open fail");
        }
    }

    /**
     * 提交事务，异常抛出
     *
     * @throws DustDbException
     */
    @Override
    public void commit() throws DustDbException {
        if (connection == null) {
            return;
        }
        try {
            connection.commit();
        } catch (SQLException e) {
            logger.debug("commit failure", e);
            throw new DustDbException("commit failure", null, e);
        }
    }

    /**
     * 回滚事务，不考虑是否成功
     * 失败记录日志
     */
    @Override
    public void rollbackQuiet() {
        try {
            rollback();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * 回滚事务，失败报出异常
     *
     * @throws DustDbException
     */
    @Override
    public void rollback() throws DustDbException {
        if (connection == null) {
            return;
        }
        try {
            connection.rollback();
        } catch (SQLException e) {
            throw new DustDbException("rollback failure", null, e);
        }
    }

    /**
     * 关闭连接，不考虑是否成功
     * 失败记录日志
     */
    @Override
    public void closeQuiet() {
        try {
            close();
        } catch (Exception ex) {
            logger.debug("close quiet exception", ex);
        }
    }

    @Override
    public void destroy() {
        this.closeQuiet();
        this.dataSource = null;
        this.connection = null;
        if (abandon) {
            LoggerHelper.warning("独立的IDataAdapter释放");
        }
    }

    @Override
    public boolean isAbandon() {
        return abandon;
    }

    @Override
    public void setAbandon(boolean abandon) {
        this.abandon = abandon;
    }

    @Override
    public void commitAndCloseQuiet() {
        if (connection == null) {
            return;
        }

        try {
            connection.commit();
        } catch (SQLException e) {
            logger.debug("commitAndCloseQuiet exception", e);
            rollbackQuiet();
            closeQuiet();
        }
    }

    /**
     * 关闭连接，不管是否成功，都会置空connect属性
     *
     * @throws DustDbException
     */
    @Override
    public void close() throws DustDbException {
        if (connection == null) {
            return;
        }
        try {
            connection.close();
        } catch (SQLException e) {
            throw new DustDbException("close failure", null, e);
        } finally {
            connection = null;
        }
    }
}
