package top.daxian1218.dust.db.sql;

import org.apache.commons.lang3.StringUtils;
import top.daxian1218.dust.commons.data.HDataRecord;
import top.daxian1218.dust.commons.data.HDataRow;
import top.daxian1218.dust.commons.exception.DustSystemException;
import top.daxian1218.dust.commons.util.TypeUtil;
import top.daxian1218.dust.db.base.HDataCommand;

import java.sql.Connection;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.function.BiFunction;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * 数据库Sql命令，dustdb不推荐使用拼接Sql的方式进行数据操作。<br>
 * 使用SqlCommand的sql加上参数的方式来访问数据库，一是提高效率的同时，也可以防止Sql注入<br>
 * 推荐使用参数化的sql书写方式，各个数据库对参数化的SQL语句都有优化，同时也可以避免书写错误或者SQL注入的问题<br>
 *
 * @author hst
 */
public class HSqlCommand extends HDataCommand {

    public static final String PrefixOld = "old_";
    public static final String OR_OPERATION = " OR ";
    public static final String AND_OPERATION = " AND ";
    public static final String INDEX_KEY = "INDEX";
    public static final String DbValueKey = "DbValue";

    private static final Pattern patterns = Pattern.compile(":([_A-Za-z][_A-Za-z0-9]*)|#\\{([_A-Za-z0-9]+)}|\\$\\{([_A-Za-z0-9]+)}");
    private final StringBuilder where = new StringBuilder();
    private final StringBuilder sql = new StringBuilder();
    private final List<String> orderList = new ArrayList<>();
    private final List<String> innerVariables = new ArrayList<>();

    private SqlCommandType commandType;
    private boolean ignoreOrder;
    private Connection connection;

    public HSqlCommand(String sql) {
        this();
        this.sql.append(sql);
    }

    public HSqlCommand() {
        super();
        commandType = SqlCommandType.QUERY;
    }

    /**
     * CommandType分为Sql文本语句，存储过程，表名
     *
     * @return
     */
    public SqlCommandType getCommandType() {
        return commandType;
    }

    /**
     * CommandType分为Sql文本语句，存储过程，表名
     * 默认值为文本语句
     *
     * @param commandType
     */
    public void setCommandType(SqlCommandType commandType) {
        this.commandType = commandType;
    }

    public boolean isIgnoreOrder() {
        return ignoreOrder;
    }

    public void setIgnoreOrder(boolean ignoreOrder) {
        this.ignoreOrder = ignoreOrder;
    }

    public Connection getConnection() {
        return connection;
    }

    public void setConnection(Connection connection) {
        this.connection = connection;
    }


    /**
     * 获取Where条件，CommandType为TableDirect时，通常使用该操作
     *
     * @return
     */
    public String getWhere() {
        return where.toString();
    }

    public void setWhere(String value) {
        where.setLength(0);
        where.append(value);
    }

    public boolean hasWhere() {
        return where.length() > 0;
    }

    @Override
    public String getCommandText() {
        return sql.toString();
    }


    /**
     * 增加Order，可以加上排序方式
     *
     * @param order
     * @return
     */
    public HSqlCommand addOrder(String order) {
        orderList.add(order);
        return this;
    }

    public HSqlCommand resetSql() {
        sql.setLength(0);
        return this;
    }

    public HSqlCommand appendSql(String sql) {
        return appendSql(sql, true);
    }

    public HSqlCommand appendSql(String sql, boolean aroundBlank) {
        if (aroundBlank) {
            this.sql.append(SqlConstants.Blank);
            this.sql.append(sql);
            this.sql.append(SqlConstants.Blank);
        } else {
            this.sql.append(sql);
        }

        return this;
    }

    /**
     * 获取参数列表，使用作为参数占位符需配合使用
     *
     * @return
     */
    public List<Object> getDbParameters() {
        return getDbParameters(getParameters());
    }

    private List<Object> getDbParameters(HDataRecord item) {
        if (item == null) {
            return new ArrayList<>();
        }

        if (commandType == SqlCommandType.PROGRAM) {
            List<Object> ps = new ArrayList<>();
            item.forEachKey(k -> {
                if (item.getValue(k) instanceof ProgramParameter) {
                    ps.add(item.getValue(k));
                }
            });
            return ps;
        }

        List<Object> paras = new ArrayList<>();
        innerVariables.forEach(innerVar -> paras.add(item.getValue(innerVar)));
        return paras;
    }

    private String getRegexKey(Matcher m) {
        char firstCh = m.group().charAt(0);
        String key = null;
        if (firstCh == ':') {
            key = m.group(1);
        } else if (firstCh == '#') {
            key = m.group(2);
        } else if (firstCh == '$') {
            key = m.group(3);
        }

        return key;
    }

    /**
     * 排序字符串，不含Order By等关键字
     *
     * @return 排序字符串
     */
    public String getOrder() {
        StringBuilder sb = new StringBuilder();
        for (String anOrderList : orderList) {
            if (StringUtils.isEmpty(anOrderList)) {
                continue;
            }
            if (sb.length() > 0) {
                sb.append(",");
            }
            sb.append(anOrderList);
        }

        return sb.toString();
    }

    /**
     * 生成执行的Sql语句
     *
     * @return 参数化的sql
     */
    protected String combineSql() {
        StringBuilder desSb = new StringBuilder(getCommandText());
        String wh = getWhere();
        if (!StringUtils.isEmpty(wh)) {
            desSb.append(" WHERE ");
            desSb.append(wh);
        }

        if (!ignoreOrder) {
            String orderStr = getOrder();
            if (!StringUtils.isEmpty(orderStr)) {
                desSb.append(" ORDER BY ");
                desSb.append(orderStr);
            }
        }

        return desSb.toString();

    }

    @Override
    public String getExecuteText() {
        return getDbExecuteSql((v, i) ->  "'" + TypeUtil.toString(v) + "'");
    }

    public String getDbExecuteSql(BiFunction<Object, Integer, String> sqlFunc) {
        String str = combineSql();
        if (!hasParameter()) {
            return str;
        }

        Matcher m = patterns.matcher(str);
        innerVariables.clear();
        int idx = 0;
        while (m.find()) {
            String key = getRegexKey(m);

            if (INDEX_KEY.equals(key)) {
                key = PrefixIndexKey + idx;
            }

            if (getParameters().containsKey(key)) {
                Object temp = getParameters().getValue(key);
                String dbValue = temp == null ? "NULL" : "'" + temp + "'";
                if (sqlFunc != null) {
                    dbValue = sqlFunc.apply(temp, idx);
                }

                str = str.replaceFirst(getRegExGroup(m.group()), dbValue);
                innerVariables.add(key);
                idx++;
            }
        }
        return str;
    }

    public HSqlCommand append(HSqlCommand whereCmd, boolean useOrOperator) {
        if (whereCmd == null ) {
            return this;
        }

        String where = whereCmd.getWhere();
        if (where.contains(OR_OPERATION)) {
            where = "(" + where + ")";
        }

        appendWhere(where, useOrOperator);
        appendParameters(whereCmd.getParameters());

        orderList.addAll(whereCmd.orderList);
        return this;
    }

    public HSqlCommand append(HSqlCommand whereCmd) {
        return append(whereCmd, false);
    }

    /**
     * 增加排序操作
     * 当sql中不存在Where时，可以使用该方法
     * strOrder可以是多个排序字段，也可以携带desc和asc
     *
     * @param strOrder
     * @return
     */
    public HSqlCommand appendOrderString(String strOrder) {
        String[] strOrders = strOrder.split(",");
        this.orderList.addAll(Arrays.asList(strOrders));
        return this;
    }

    public HSqlCommand appendWhere(String sql) {
        return appendWhere(sql, false);
    }


    /**
     * 语句的Where条件
     * <ul>
     * <li>当sql中不存在Where时，可以使用该方法</li>
     * <li>支持OR拼接</li>
     * </ul>
     * OR操作如果当前没有其他Where条件，则会累积到下一个条件
     * <code>
     * // cmd.where = ""
     * cmd.appendWhere("a=1", true);
     * // cmd.where = "a=1"
     * cmd.appWhere("b=2", false);
     * //cmd.where = "a=1 OR b=2"
     * </code>
     *
     * @param sql
     * @param useOrOperator
     * @return
     */
    public HSqlCommand appendWhere(String sql, Boolean useOrOperator) {
        if (StringUtils.isEmpty(sql)) {
            return this;
        }

        if (where.length() > 0) {
            where.append(useOrOperator ? OR_OPERATION : AND_OPERATION);
        }

        where.append(sql);
        return this;
    }

    private String getRegExGroup(String srcGroup) {
        return Pattern.quote(srcGroup);
    }

    public Integer getBeginIndex() {
        return this.getPageIndex() * this.getPageSize();
    }

    public Integer getEndIndex() {
        return (this.getPageIndex() + 1) * this.getPageSize() - 1;
    }

    public ProgramParameter getProgramParameter(String name) {
        ProgramParameter p = (ProgramParameter) getParameter(name);
        if (p == null) {
            throw new DustSystemException("type of parameter is not store proc parameter");
        }

        return p;
    }
}
