package top.daxian1218.dust.db.sql;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import top.daxian1218.dust.commons.data.HDataColumn;
import top.daxian1218.dust.commons.data.HDataTable;
import top.daxian1218.dust.commons.data.StatType;
import top.daxian1218.dust.commons.generic.Preconditions;
import top.daxian1218.dust.commons.log.LoggerHelper;
import top.daxian1218.dust.commons.util.TypeUtil;
import top.daxian1218.dust.db.DustDbException;
import top.daxian1218.dust.db.base.HDataCommand;
import top.daxian1218.dust.db.base.IDbOperations;
import top.daxian1218.dust.db.base.UpdateResult;
import top.daxian1218.dust.db.util.DbUtil;

import java.math.BigDecimal;
import java.sql.*;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * jdbc相关方法的封装类，主要是数据库操作封装，查询，更新，执行
 * 封装了不同关系数据库的公用方法,
 *
 * @author hst
 */
public abstract class SqlDbOperations implements IDbOperations {

    static Logger logger = LoggerFactory.getLogger(SqlDbOperations.class);

    protected final String RowStartParameterName = "start";
    protected final String RowEndParameterName = "end";

    protected HSqlCommand toSqlCommand(HDataCommand cmd) {
        if (!(cmd instanceof HSqlCommand)) {
            throw new DustDbException("cmd must be instanceof SqlCommand");
        }

        return (HSqlCommand) cmd;
    }

    public void fillStatement(PreparedStatement stmt, List<Object> params) {
        if (params == null || stmt == null) {
            return;
        }

        try {
            for (int len = params.size(), i = 0; i < len; i++) {
                Object p = params.get(i);
                int index = i + 1;
                Object dbValue = getDbValue(p, stmt.getConnection());
                if (dbValue != null) {
                    stmt.setObject(index, dbValue);
                } else {
                    stmt.setNull(index, Types.NULL);
                }
            }

        } catch (SQLException ex) {
            throw formatException(ex, null);
        }
    }

    protected Object getDbValue(Object val, Connection connection) throws SQLException {
        if (val == null) {
            return null;
        }

        if (val instanceof Date) {
            return new Timestamp(((Date) val).getTime());
        }

        if (val instanceof Boolean) {
            return (Boolean)val ? 1 : 0;
        }

        return val;
    }

    /**
     * 释放数据集和数据库请求
     *
     * @param rs
     */
    public void releaseResultSetAndStatement(ResultSet rs, Statement stmt) {
        try {
            if (null != rs) {
                rs.close();
            }
        } catch (SQLException ex) {
            LoggerHelper.debug(ex);
        }

        try {
            if (null != stmt) {
                stmt.close();
            }
        } catch (SQLException ex) {
            LoggerHelper.debug(ex);
        }

    }

    /**
     * 生成SQL，需要提供具体的分页SQL
     * @param cmd
     * @return
     */
    protected abstract String getSql(HSqlCommand cmd, boolean isQuery);

    public PreparedStatement getQueryStatement(HSqlCommand cmd) throws SQLException {
        Connection conn = cmd.getConnection();
        if (conn == null) {
            throw new DustDbException("connection is null");
        }

        PreparedStatement stmt = conn.prepareStatement(getSql(cmd, true));
        List<Object> dbParams = cmd.getDbParameters();
        if (cmd.getPageSize() > 0) {
            aggregateData(cmd);
            setPageParameter(dbParams, cmd);
        }

        fillStatement(stmt, dbParams);
        return stmt;
    }

    public void setPageParameter(List<Object> parameters, HSqlCommand cmd) {
        parameters.add(cmd.getBeginIndex() + 1);
        parameters.add(cmd.getEndIndex() + 1);
    }

    protected DustDbException formatException(SQLException ex, HSqlCommand cmd) {
        return DustDbException.fromDbException(ex, cmd);
    }

    public PreparedStatement getUpdateStatement(HSqlCommand cmd) {
        Connection conn = cmd.getConnection();
        if (conn == null) {
            throw new DustDbException("connection is null");
        }

        PreparedStatement stmt = null;
        try {
            stmt = conn.prepareStatement(getSql(cmd, false));
        } catch (SQLException sqe) {
            throw formatException(sqe, cmd);
        }

        PreparedStatement finalStmt = stmt;
        cmd.iterate(it -> {
            fillStatement(finalStmt, ((HSqlCommand) it).getDbParameters());
            try {
                finalStmt.addBatch();
            } catch (SQLException sqe) {
                throw formatException(sqe, cmd);
            }

            return true;
        });
        return stmt;
    }

    /**
     * 根据cmd，执行查询操作，返回结果集DataTable
     * <strong>注意：</strong>非Select操作，请勿调用该方法，有些操作可能导致异常
     * 存储过程返回游标的操作也可用该方法
     *
     * @param cmd
     * @return
     * @throws DustDbException
     */
    public <T> T query(HDataCommand cmd, Function<ResultSet, T> readerHandler) throws DustDbException {
        HSqlCommand sqlCmd = (HSqlCommand) cmd;
        if (sqlCmd == null) {
            throw new DustDbException("非法的SQL命令");
        }

        DbUtil.printSql(sqlCmd);
        if (sqlCmd.getCommandType() == SqlCommandType.PROGRAM) {
            return executeProcedure(sqlCmd, readerHandler);
        }

        PreparedStatement statement = null;
        ResultSet rs = null;
        try {
            statement = getQueryStatement(sqlCmd);
            rs = statement.executeQuery();
            return readerHandler.apply(rs);
        } catch (SQLException e) {
            throw formatException(e, sqlCmd);
        } finally {
            releaseResultSetAndStatement(rs, statement);
        }
    }

    /**
     * 执行存储过程，执行结果，请从参数cmd中取
     *
     * @param cmd
     * @throws DustDbException
     * @see HSqlCommand#getParameters() ()
     * @see HDataCommand#getParameter(String)
     */
    protected <T> T executeProcedure(HSqlCommand cmd, Function<ResultSet, T> readerHandler) throws DustDbException {
        Preconditions.checkNotNull(readerHandler);

        CallableStatement cs = null;
        ResultSet rs = null;
        try {
            List<ProgramParameter> params = cmd.getDbParameters().stream()
                    .map(p -> (ProgramParameter) p).collect(Collectors.toList());
            cs = cmd.getConnection().prepareCall(cmd.getDbExecuteSql((val, index) -> "?"));
            fillProcParams(cs, params);
            boolean hasResult = cs.execute();
            T result = null;
            if (hasResult) {
                rs = cs.getResultSet();
                result = readerHandler.apply(rs);
            }

            // 处理输出参数
            List<HDataTable> ds = new ArrayList<>();
            for (int i = 0; i < params.size(); i++) {
                ProgramParameter param = params.get(i);
                if (param.getParamIoType() == ProcParaType.INPUT) {
                    continue;
                }
                // 游标,数据集返回参数
                if (SqlDataType.CURSOR == param.getParamDataType()) {
                    if (hasResult) {
                        param.setValue(result);
                    } else {
                        result = readerHandler.apply((ResultSet) cs.getObject(i + 1));
                        param.setValue(result);
                    }

                } else {
                    Object csTmp = cs.getObject(i + 1);
                    param.setValue(csTmp);
                }
            }

            return result;
        } catch (SQLException e) {
            throw formatException(e, cmd);
        } finally {
            releaseResultSetAndStatement(rs, cs);
        }
    }

    /**
     * 更新数据，可执行insert, update, delete,以及有关数据库结构的操作（需要相应的权限）
     *
     * @param cmd
     * @return
     * @throws DustDbException
     */
    public UpdateResult update(HSqlCommand cmd) throws DustDbException {
        Connection conn = cmd.getConnection();
        if (conn == null) {
            throw new DustDbException("connection is null");
        }

        switch (cmd.getCommandType()) {
            case PROGRAM: {
                executeProcedure(cmd, resultSet -> null);
                return new UpdateResult(0);
            }
            case TEXT: {
                try {
                    try (Statement statement = cmd.getConnection().createStatement()) {
                        statement.execute(cmd.getCommandText());
                    }
                    return new UpdateResult(0);
                } catch (SQLException sqe) {
                    throw formatException(sqe, cmd);
                }
            }
            default: {
                PreparedStatement stmt = getUpdateStatement(cmd);
                int[] rows;
                try {
                    rows = stmt.executeBatch();
                } catch (SQLException ex) {
                    throw formatException(ex, cmd);
                } finally {
                    releaseResultSetAndStatement(null, stmt);
                }
                return new UpdateResult(rows);
            }
        }
    }

    @Override
    public boolean isSupport(String dbType) {
        return true;
    }

    protected void fillProcParams(CallableStatement cs, List<ProgramParameter> params) throws SQLException {
        for (int i = 0, size = params.size(); i < size; i++) {
            ProgramParameter param = params.get(i);
            if (param.getParamIoType() != ProcParaType.INPUT) {
                cs.registerOutParameter(i + 1, getSqlType(param));
            }
            if (param.getParamIoType() == ProcParaType.INPUT
                    || param.getParamIoType() == ProcParaType.INOUT) {
                Object v = getDbValue(param.getValue(), cs.getConnection());
                if (v != null) {
                    cs.setObject(i + 1, v);
                } else {
                    int sqlType = getSqlType(param);
                    cs.setNull(i + 1, sqlType);
                }
            }
        }
    }

    protected int getSqlType(ProgramParameter p) {
        switch (p.getParamDataType()) {
            case INTEGER:
                return Types.INTEGER;
            case NUMBER:
            case TIMESTAMP:
                return Types.NUMERIC;
            case DATETIME:
                return Types.DATE;
            case STRING:
                return Types.NVARCHAR;
            case CLOB:
                return Types.CLOB;
            case CURSOR:
                return Types.REF_CURSOR;
            case BLOB:
                return Types.BLOB;
            default:
                return Types.VARCHAR;
        }
    }

    protected void aggregateData(HSqlCommand cmd) throws DustDbException {

        ResultSet rs = null;
        PreparedStatement statement = null;

        try {
            StringBuilder execSql = new StringBuilder("select count(*) as nTotal");
            cmd.getFields().forEach(col -> {
                if (col.getStatType() == StatType.NONE || col.getStatType() == StatType.COUNT) { return; }

                execSql.append(",");
                execSql.append(getAggregateSql(col));
            });

            execSql.append(" from (");
            execSql.append(cmd.getDbExecuteSql((val, idx) -> "?"));
            execSql.append(") row_ ");
            statement = cmd.getConnection().prepareStatement(execSql.toString());
            fillStatement(statement, cmd.getDbParameters());
            rs = statement.executeQuery();
            if (rs.next()) {
                cmd.setTotalRows(rs.getInt(1));
                ResultSet finalRs = rs;
                cmd.getFields().forEach(col -> {
                    if (col.getStatType() == StatType.NONE) {
                        return;
                    }

                    if (col.getStatType() == StatType.COUNT) {
                        col.setStatValue(TypeUtil.toDecimal(cmd.getTotalRows()));
                        return;
                    }

                    try {
                        col.setStatValue(TypeUtil.toDecimal(finalRs.getBigDecimal(col.getName())));
                    } catch (SQLException e) {
                       logger.error("列：" + col.getName() + ",获取到无效的统计值", e);
                    }
                });
            }
        } catch (SQLException e) {
            throw formatException(e, cmd);
        } finally {
            releaseResultSetAndStatement(rs, statement);
        }
    }

    protected String getAggregateSql(HDataColumn col) {
        switch (col.getStatType()) {
            case AVG:
                return "avg(" + col.getName() + ") as " + col.getName();
            case MAX:
                return "max(" + col.getName() + ") as " + col.getName();
            case MIN:
                return "min(" + col.getName() + ") as " + col.getName();
            case SUM:
                return "sum(" + col.getName() + ") as " + col.getName();
        }
        return null;
    }


    /**
     * 更新数据，支持修改，新增和删除操作
     * @param cmd
     * @return
     * @throws DustDbException
     */
    @Override
    public UpdateResult update(HDataCommand cmd) throws DustDbException {
        HSqlCommand sqlCommand = toSqlCommand(cmd);
        return update(sqlCommand);
    }
}
