package top.daxian1218.dust.db.sql.pool;

import com.alibaba.druid.pool.DruidAbstractDataSource;
import com.alibaba.druid.pool.DruidDataSource;
import top.daxian1218.dust.commons.exception.DustSystemException;
import top.daxian1218.dust.commons.log.LoggerHelper;
import top.daxian1218.dust.db.base.DbAccess;
import top.daxian1218.dust.commons.util.NameUtil;
import top.daxian1218.dust.commons.util.ClassUtil;
import top.daxian1218.dust.db.base.DbTypes;
import top.daxian1218.dust.db.sql.IDataSource;
import org.apache.commons.lang3.StringUtils;
import top.daxian1218.dust.db.util.DbUtil;

import java.lang.reflect.Method;
import java.sql.SQLException;


/**
 * Druid连接池的Template类
 * <p>
 * 实现了根据DbContext创建新的数据库连接池的方法
 * </p>
 *
 * @author hst
 */
public class DruidTemplate implements IDataSourceTemplate {

    private DataSourceProperties dataSourceProperties = new DataSourceProperties();

    public DruidTemplate() {
        dataSourceProperties.put("initialSize", 2);
        dataSourceProperties.put("maxActive", 600);
        dataSourceProperties.put("minIdle", 2);
        dataSourceProperties.put("maxWait", 10000);
        dataSourceProperties.put("poolPreparedStatements", true);
        dataSourceProperties.put("maxPoolPreparedStatementPerConnectionSize", 33);
        dataSourceProperties.put("testOnBorrow", false);
        dataSourceProperties.put("testOnReturn", false);
        dataSourceProperties.put("testWhileIdle", true);
        dataSourceProperties.put("timeBetweenEvictionRunsMillis", 60000);
        dataSourceProperties.put("minEvictableIdleTimeMillis", 25200000);
        dataSourceProperties.put("removeAbandoned", true);
        dataSourceProperties.put("logAbandoned", true);
        dataSourceProperties.put("filters", "stat");
        dataSourceProperties.put("validationQuery", "Select 1");
    }

    public DruidTemplate(DataSourceProperties dataSourceProperties) {
        this();
        this.dataSourceProperties.putAll(dataSourceProperties);
    }


    public DataSourceProperties getDataSourceProperties() {
        return dataSourceProperties;
    }

    public void setDataSourceProperties(DataSourceProperties dataSourceProperties) {
        this.dataSourceProperties = dataSourceProperties;
    }


    /**
     * 创建一个连接池数据源，加入到druidDataSources缓存集合
     * <strong>注意：</strong>如果缓存集合已经存在连接信息完全一致的数据源，则从缓存读取
     * 如果initWhenCreate为true则会调用初始化操作
     *
     * @param db
     * @return
     */
    @Override
    public IDataSource createDataSource(DbAccess db) throws SQLException {
        DruidDataSource ds = createInstance();
        ds.setUrl(DbUtil.getUrl(db.getDbType(), db.getHost(), db.getDbName(), db.getQueryString()));
        ds.setUsername(db.getUser());
        ds.setPassword(db.getPassword());
        ds.setName(db.getName());
        ds.setDriverClassName(getDriverClassName(db.getDbType()));

        if (!StringUtils.isEmpty(db.getValidationQuery())) {
            ds.setValidationQuery(db.getValidationQuery());
        } else {
            ds.setValidationQuery(getDefaultValidationQuery(db.getDbType()));
        }

        ds.init();
        return new DustDataSource(db, ds);
    }

    private String getDriverClassName(String dbType) {
        switch (dbType) {
            case DbTypes.ClickHouse: {
                return "com.github.housepower.jdbc.ClickHouseDriver";
            }
            case DbTypes.Highgo:{
                return "com.highgo.jdbc.Driver";
            }
            default: {
                return StringUtils.EMPTY;
            }
        }
    }

    private String getDefaultValidationQuery(String dbType) {
        switch (dbType) {
            case DbTypes.Oracle:
                return "SELECT SYSDATE FROM DUAL";
            default:
                return "SELECT 1";
        }
    }

    private DruidDataSource createInstance() throws SQLException {
        try {
            DruidDataSource ds = new DruidDataSource();
            Method[] methods = DruidAbstractDataSource.class.getMethods();
            for (Method method : methods) {
                String key = fieldOfMethod(method);
                if (key != null && dataSourceProperties.containsKey(key)) {
                    method.invoke(ds, ClassUtil.typeConvert(dataSourceProperties.get(key), method.getParameterTypes()[0]));
                }
            }

            return ds;
        } catch (Exception ex) {
            throw new SQLException("", ex);
        }
    }

    private String fieldOfMethod(Method method) {
        if (method.getParameterTypes().length == 1) {
            String key = method.getName();
            if (key.startsWith("set") && key.length() > 3) {
                key = key.substring(3);
                return NameUtil.toLowerCaseFirstLetter(key);
            }

            if (key.startsWith("is") && key.length() > 2) {
                key = key.substring(2);
                return NameUtil.toLowerCaseFirstLetter(key);
            }
        }

        return null;
    }
}
