package top.daxian1218.dust.db.support.mssql;

import org.apache.commons.lang3.StringUtils;
import top.daxian1218.dust.commons.data.HDataColumn;
import top.daxian1218.dust.commons.data.HDataTable;
import top.daxian1218.dust.commons.generic.Preconditions;
import top.daxian1218.dust.commons.util.TypeUtil;
import top.daxian1218.dust.db.base.DbTypes;
import top.daxian1218.dust.db.base.IDataAdapter;
import top.daxian1218.dust.db.dict.*;
import top.daxian1218.dust.db.sql.HSqlCommand;
import top.daxian1218.dust.db.sql.command.HInsertSqlCommand;
import top.daxian1218.dust.db.sql.command.HUpdateSqlCommand;

/**
 * @author hst
 */
public class MssqlBuilder extends SqlBuilder {

    @Override
    public boolean isSupport(String dbType) {
        return StringUtils.equals(dbType, DbTypes.Mssql);
    }


    public String getTableScript(DictObj destObj) {
        Preconditions.checkNotNull(destObj);

        StringBuilder sb = new StringBuilder();
        sb.append("CREATE TABLE [" + destObj.getTableName() + "] (");
        sb.append("\r\n");

        StringBuilder primaryKeys = new StringBuilder();
        destObj.getColumns().forEach(col ->
        {
            if (!StringUtils.equals(col.getTableName(), destObj.getTableName())) {
                return;
            }

            if (sb.charAt(sb.length() - 1) != '\n') {
                sb.append(",");
                sb.append("\r\n");
            }

            sb.append(getDbColumnType(col));

            if (col.isPrimaryKey()) {
                if (primaryKeys.length() > 0) {
                    primaryKeys.append(",");
                }

                primaryKeys.append("[" + col.getPhysicalName() + "]");
            }
        });

        sb.append(",\r\n");
        sb.append("  PRIMARY KEY CLUSTERED  (");
        sb.append(primaryKeys);
        sb.append(")");

        sb.append("\r\n");
        sb.append(") ");

        return sb.toString();
    }

    private String getDbColumnType(DictObjColumn col) {
        StringBuilder sbCol = new StringBuilder();
        sbCol.append("  [");
        sbCol.append(col.getPhysicalName());
        sbCol.append("] ");
        switch (col.getDataType()) {
            case BINARY:
                sbCol.append("[image]");
                break;
            case TEXT:
                if (col.getDataWidth() < 5000 && col.getDataWidth() > 0) {
                    sbCol.append("[nvarchar](" + col.getDataWidth() + ")");
                } else {
                    sbCol.append("[text]");
                }
                break;
            case STRING:
                sbCol.append("[nvarchar](" + col.getDataWidth() + ")");
                break;
            case INT:
                sbCol.append("int");
                break;
            case UBIGINT:
                sbCol.append("bigint");
                break;
            case BOOLEAN:
                sbCol.append("tinyint");
                break;
            case CURRENCY:
            case NUMBER:
                sbCol.append("[numeric](" + col.getDataWidth() + "," + col.getDataDec() + ")");
                break;
            case DATE:
                sbCol.append("datetime");
                break;
        }

        if (col.isRequired()) {
            sbCol.append(" NOT NULL");
        } else if (StringUtils.isEmpty(col.getDefaultValue())) {
            sbCol.append(" NULL");
        }

        if (!StringUtils.isEmpty(col.getDefaultValue())) {
            if (col.getDataType() == DataType.DATE
                    && StringUtils.equals(col.getDefaultValue(), DictConstant.DEFAULT_VALUE_NOW)) {
                sbCol.append(" DEFAULT GetDate()");
            } else if (col.getDataType() == DataType.DATE
                    && StringUtils.equals(col.getDefaultValue(), DictConstant.DEFAULT_VALUE_UPDATE)) {
                sbCol.append(" DEFAULT GetDate()");
            } else {
                sbCol.append(" DEFAULT ");
                sbCol.append("(");
                sbCol.append(col.getDefaultValue());
                sbCol.append(")");
            }

        }

        if (col.getDataType() == DataType.UBIGINT && col.isAutoIncrement()) {
            sbCol.append(" IDENTITY(1,1)");
        }

        return sbCol.toString();
    }

    @Override
    protected String getDbValue(Object o) {
        if (o == null) {
            return "NULL";
        }

        switch (o.getClass().getSimpleName()) {
            case "Integer":
            case "int":
            case "Long":
            case "long":
            case "Double":
            case "double":
            case "Float":
            case "float":
            case "BigDecimal":
                return TypeUtil.toString(o);
            case "String":
                return "'" + o + "'";
            case "Boolean":
            case "boolean":
                return TypeUtil.toBoolean(o) ? "1" : "0";
            case "Date":
                return "CAST(N'" + TypeUtil.toString(o, "yyyy-MM-dd HH:mm:ss") + "' AS DateTime)";
        }

        return TypeUtil.toString(o);
    }

    @Override
    protected long getLastId(IDataAdapter adapter, DictObj destObj, String incrementColumn) {
        HSqlCommand cmd = new HSqlCommand("SELECT @@IDENTITY as ID");
        HDataTable dt = adapter.query(cmd);
        if (dt.size() == 1) {
            return TypeUtil.toLong(dt.getScalar());
        }
        return 0;
    }

    @Override
    protected boolean handleUpdateSqlCoreColumn(DictObjColumn col, HUpdateSqlCommand command) {
        if (super.handleUpdateSqlCoreColumn(col, command)) {
            return true;
        }

        if (StringUtils.equals(col.getDefaultValue(), DictConstant.DEFAULT_VALUE_UPDATE)
                && col.getDataType() == DataType.DATE) {
            HDataColumn dc = new HDataColumn(col.getPhysicalName());
            dc.setIgnore(true);
            dc.getProperties().set(HSqlCommand.DbValueKey, "GETDATE()");
            command.getFields().add(dc);
            return true;
        }

        return false;
    }

    @Override
    protected boolean handleInsertSqlCoreColumn(DictObjColumn col, HInsertSqlCommand command) {
        if (col.isAutoIncrement() || col.isIgnore() || StringUtils.isEmpty(col.getPhysicalName())) {
            return true;
        }

        boolean dbDate = col.getDataType() == DataType.DATE &&
                (StringUtils.equals(col.getDefaultValue(), DictConstant.DEFAULT_VALUE_UPDATE) ||
                        StringUtils.equals(col.getDefaultValue(), DictConstant.DEFAULT_VALUE_UPDATE));
        if (dbDate) {
            HDataColumn dc = new HDataColumn(col.getPhysicalName());
            dc.setIgnore(true);
            dc.getProperties().set(HSqlCommand.DbValueKey, "GETDATE()");
            command.getFields().add(dc);
            return true;
        }

        return false;
    }
}
