package top.daxian1218.dust.db.support.mysql;

import top.daxian1218.dust.commons.util.TypeUtil;
import top.daxian1218.dust.db.DustDbException;
import top.daxian1218.dust.db.base.DbTypes;
import top.daxian1218.dust.db.dict.*;
import top.daxian1218.dust.commons.data.HDataTable;
import top.daxian1218.dust.db.base.IDataAdapter;
import top.daxian1218.dust.db.sql.HSqlCommand;
import org.apache.commons.lang3.StringUtils;

import static top.daxian1218.dust.commons.generic.Preconditions.checkNotNull;
import static top.daxian1218.dust.db.dict.DictConstant.DEFAULT_VALUE_NOW;
import static top.daxian1218.dust.db.dict.DictConstant.DEFAULT_VALUE_UPDATE;

/**
 * @author hst
 */
public class MySqlBuilder extends SqlBuilder {

    private boolean checkIdRelationColumn(DictObj destObj, String tbName) {
        for (DictObjColumn col : destObj.getColumns()) {
            if (StringUtils.isNotEmpty(col.getColumnName()) && StringUtils.equals(col.getTableName(), tbName)) {
                return true;
            }
        }
        
        return false;
    }


    @Override
    public long getLastId(IDataAdapter adapter, DictObj destObj, String incrementColumn) throws DustDbException {
        HSqlCommand cmd = new HSqlCommand("SELECT @@IDENTITY as ID");
        HDataTable dt = adapter.query(cmd);
        if (dt.getRows().size() == 1) {
            return TypeUtil.toInteger(dt.getRows().get(0).get("ID"));
        }
        return 0;
    }

    @Override
    public boolean isSupport(String dbType) {
        return StringUtils.equals(dbType, DbTypes.Mysql);
    }


    public String getTableScript(DictObj destObj) {
        checkNotNull(destObj);

        StringBuilder sb = new StringBuilder();
        sb.append("CREATE TABLE `");
        sb.append(destObj.getTableName());
        sb.append("` (");
        sb.append("\r\n");

        StringBuilder primaryKeys = new StringBuilder();
        for (int i = 0; i < destObj.getColumns().size(); i++) {
            DictObjColumn col = destObj.getColumn(i);
            if (!StringUtils.equals(col.getTableName(), destObj.getTableName())) {
                continue;
            }

            if (sb.charAt(sb.length() - 1) != '\n') {
                sb.append(",");
                sb.append("\r\n");
            }

            sb.append(getDbColumnType(col));

            if (col.isPrimaryKey()) {
                if (primaryKeys.length() > 0) {
                    primaryKeys.append(",");
                }

                primaryKeys.append("`" + col.getColumnName() + "`");
            }
        }

        sb.append(",\r\n");
        sb.append("  PRIMARY KEY (");
        sb.append(primaryKeys);
        sb.append(")");

        sb.append("\r\n");
        sb.append(") ENGINE=InnoDB DEFAULT CHARSET=utf8;");

        return sb.toString();
    }

    private String getDbColumnType(DictObjColumn col) {
        StringBuilder sbCol = new StringBuilder();
        sbCol.append("  `");
        sbCol.append(col.getColumnName());
        sbCol.append("` ");
        switch (col.getDataType()) {
            case BINARY:
                sbCol.append("blob");
                break;
            case TEXT:
                if (col.getDataWidth() < 5000 && col.getDataWidth() > 0) {
                    sbCol.append("varchar(");
                    sbCol.append(col.getDataWidth());
                    sbCol.append(")");
                } else {
                    sbCol.append("text");
                }
                break;
            case STRING:
                if (col.getDataWidth() > 255) {
                    sbCol.append("varchar(");
                } else {
                    sbCol.append("char(");
                }

                sbCol.append(col.getDataWidth());
                sbCol.append(")");
                break;
            case INT:
                sbCol.append("int");
                break;
            case UBIGINT:
                sbCol.append("bigint unsigned");
                break;
            case BOOLEAN:
                sbCol.append("tinyint unsigned");
                break;
            case CURRENCY:
            case NUMBER:
                sbCol.append("numeric(");
                sbCol.append(col.getDataWidth());
                sbCol.append(",");
                sbCol.append(col.getDataDec());
                sbCol.append(")");
                break;
            case DATE:
                sbCol.append("datetime");
                break;
            default:
                sbCol.append(StringUtils.EMPTY);
                break;
        }

        if (col.isRequired()) {
            sbCol.append(" NOT NULL");
        } else if (StringUtils.isEmpty(col.getDefaultValue())) {
            sbCol.append(" DEFAULT NULL");
        }

        if (!StringUtils.isEmpty(col.getDefaultValue())) {
            if (col.getDataType() == DataType.DATE && StringUtils.equals(col.getDefaultValue(), DEFAULT_VALUE_NOW)) {
                sbCol.append(" DEFAULT CURRENT_TIMESTAMP");
            } else if (col.getDataType() == DataType.DATE && StringUtils.equals(col.getDefaultValue(), DEFAULT_VALUE_UPDATE)) {
                sbCol.append(" DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP");
            } else {
                sbCol.append(" DEFAULT ");
                sbCol.append("'");
                sbCol.append(col.getDefaultValue());
                sbCol.append("'");
            }

        }

        if (col.getDataType() == DataType.UBIGINT && col.isAutoIncrement()) {
            sbCol.append(" AUTO_INCREMENT");
        }

        return sbCol.toString();
    }

    @Override
    protected String getDbValue(Object o) {
        if (o == null) {
            return "NULL";
        }

        switch (o.getClass().getSimpleName()) {
            case "Integer":
            case "int":
            case "Long":
            case "long":
            case "Short":
            case "short":
            case "Double":
            case "double":
            case "BigDecimal":
            case "Float":
            case "float":
                return TypeUtil.toString(o);
            case "String":
                return "'" + o + "'";
            case "Boolean":
            case "boolean":
                return TypeUtil.toBoolean(o) == true ? "1" : "0";
            case "Date":
                return "str_to_date('" + TypeUtil.toString(o, "yyyy-MM-dd HH:mm:ss") + "','%Y-%m-%d %Y-%m-%d %r')";
        }

        return TypeUtil.toString(o);
    }
}
