package top.daxian1218.dust.db.support.oracle;

import org.apache.commons.lang3.StringUtils;
import top.daxian1218.dust.commons.data.HDataColumn;
import top.daxian1218.dust.commons.util.TypeUtil;
import top.daxian1218.dust.db.base.DbTypes;
import top.daxian1218.dust.db.base.IDataAdapter;
import top.daxian1218.dust.db.dict.*;
import top.daxian1218.dust.db.sql.HSqlCommand;
import top.daxian1218.dust.db.sql.command.HInsertSqlCommand;

import static top.daxian1218.dust.commons.generic.Preconditions.checkNotNull;
import static top.daxian1218.dust.db.dict.DictConstant.DEFAULT_VALUE_NOW;
import static top.daxian1218.dust.db.dict.DictConstant.DEFAULT_VALUE_UPDATE;

/**
 * Sql Builder for Oracle
 *
 * @author hst
 */
public class OracleBuilder extends SqlBuilder {


    @Override
    protected long getLastId(IDataAdapter adapter, DictObj destObj, String incrementColumn) {
        return -1;
    }

    @Override
    public boolean isSupport(String dbType) {
        return StringUtils.equals(dbType, DbTypes.Oracle);
    }


    public String getTableScript(DictObj destObj) {
        checkNotNull(destObj);

        StringBuilder sb = new StringBuilder();
        sb.append("CREATE TABLE ");
        sb.append(destObj.getTableName());
        sb.append(" (");
        sb.append("\r\n");

        StringBuilder primaryKeys = new StringBuilder();
        for (int i = 0; i < destObj.getColumns().size(); i++) {
            DictObjColumn col = destObj.getColumn(i);
            if (!StringUtils.equals(col.getTableName(), destObj.getTableName())) {
                continue;
            }

            if (sb.charAt(sb.length() - 1) != '\n') {
                sb.append(",");
                sb.append("\r\n");
            }

            sb.append(getDbColumnType(col));

            if (col.isPrimaryKey()) {
                if (primaryKeys.length() > 0) {
                    primaryKeys.append(",");
                }

                primaryKeys.append(col.getColumnName());
            }
        }

        sb.append("\r\n");
        sb.append(");");

        //主键
        sb.append("ALTER TABLE");
        sb.append(destObj.getTableName());
        sb.append("\r\n");
        sb.append("  ADD CONSTRAINT");
        sb.append(" PK_" + destObj.getTableName());
        sb.append(" PRIMARY KEY (");
        sb.append(primaryKeys);
        sb.append(");\r\n");

        //序列
        sb.append("\r\n");
        sb.append("CREATE SEQUENCE ");
        sb.append("SEQ_" + destObj.getTableName());
        sb.append(" minvalue 1 maxvalue 999999999999999999999999999 start with 1 increment by 1 cache 20 order;");
        return sb.toString();
    }

    private String getDbColumnType(DictObjColumn col) {
        StringBuilder sbCol = new StringBuilder();
        sbCol.append(" ");
        sbCol.append(col.getColumnName());
        sbCol.append(" ");
        switch (col.getDataType()) {
            case BINARY:
                sbCol.append("blob");
                break;
            case TEXT:
                sbCol.append("clob");
                break;
            case STRING:
                if (col.getDataWidth() < 2000) {
                    sbCol.append("varchar2(");
                    sbCol.append(col.getDataWidth());
                    sbCol.append(")");
                } else {
                    sbCol.append("clob");
                }
                break;
            case INT:
                sbCol.append("number(19,0)");
                break;
            case UBIGINT:
                sbCol.append("number(20,0)");
                break;
            case BOOLEAN:
                sbCol.append("number(1,0)");
                break;
            case CURRENCY:
            case NUMBER:
                sbCol.append("number(");
                sbCol.append(col.getDataWidth());
                sbCol.append(",");
                sbCol.append(col.getDataDec());
                sbCol.append(")");
                break;
            case DATE:
                sbCol.append("date");
                break;
            default:
                sbCol.append(StringUtils.EMPTY);
                break;
        }


        if (!StringUtils.isEmpty(col.getDefaultValue())) {
            if (col.getDataType() == DataType.DATE && StringUtils.equals(col.getDefaultValue(), DEFAULT_VALUE_NOW)) {
                sbCol.append(" DEFAULT sysdate");
            } else if (col.getDataType() == DataType.DATE && StringUtils.equals(col.getDefaultValue(), DEFAULT_VALUE_UPDATE)) {
                sbCol.append(" DEFAULT sysdate");
            } else {
                sbCol.append(" DEFAULT ");
                sbCol.append("'");
                sbCol.append(col.getDefaultValue());
                sbCol.append("'");
            }

        }

        if (col.isRequired()) {
            sbCol.append(" NOT NULL");
        }

        return sbCol.toString();
    }


    @Override
    protected String getDbValue(Object o) {
        if (o == null) {
            return "NULL";
        }

        switch (o.getClass().getSimpleName()) {
            case "Integer":
            case "int":
            case "Long":
            case "long":
            case "Short":
            case "short":
            case "Double":
            case "double":
            case "BigDecimal":
            case "Float":
            case "float":
                return TypeUtil.toString(o);
            case "String":
                return "'" + o + "'";
            case "Boolean":
            case "boolean":
                return TypeUtil.toBoolean(o) == true ? "1" : "0";
            case "Date":
                return "to_date('" + TypeUtil.toString(o, "yyyy-MM-dd HH:mm:ss") + "','yyyy-mm-dd hh24:mi:ss')";
            default:
                return TypeUtil.toString(o);
        }
    }

    @Override
    protected boolean handleInsertSqlCoreColumn(DictObjColumn col, HInsertSqlCommand command) {
        if (col.isAutoIncrement() || col.isIgnore() || StringUtils.isEmpty(col.getPhysicalName())) {
            return true;
        }

        boolean dbDate = col.getDataType() == DataType.DATE &&
                (StringUtils.equals(col.getDefaultValue(), DictConstant.DEFAULT_VALUE_UPDATE) ||
                        StringUtils.equals(col.getDefaultValue(), DictConstant.DEFAULT_VALUE_UPDATE));
        if (dbDate) {
            HDataColumn dc = new HDataColumn(col.getPhysicalName());
            dc.setIgnore(true);
            dc.getProperties().set(HSqlCommand.DbValueKey, "sysdate");
            command.getFields().add(dc);
            return true;
        }

        return false;
    }
}
