package top.daxian1218.dust.db.util;

import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import top.daxian1218.dust.commons.data.HDataColumn;
import top.daxian1218.dust.commons.data.HDataRow;
import top.daxian1218.dust.commons.data.HDataTable;
import top.daxian1218.dust.commons.id.SnowFlakeIdWorker;
import top.daxian1218.dust.commons.log.LoggerHelper;
import top.daxian1218.dust.commons.util.Base64Util;
import top.daxian1218.dust.commons.util.TypeUtil;
import top.daxian1218.dust.db.DbAdapterManager;
import top.daxian1218.dust.db.DictGlobalConfig;
import top.daxian1218.dust.db.DustDbException;
import top.daxian1218.dust.db.base.DbTypes;
import top.daxian1218.dust.db.base.IDataAdapter;
import top.daxian1218.dust.db.dict.DataType;
import top.daxian1218.dust.db.sql.HSqlCommand;

import java.sql.*;

/**
 * 数据库操纵工具类，如处理sql参数以及关闭相关连接等
 *
 * @author hst
 */
public class DbUtil {
    private static final String SQL_LOG = "insert into SYS_LOG_SQL (ID, MESSAGE, ERROR_SQL) VALUES (:id, :msg, :sql)";
    private static final String LOG_TAG = "LOG";

    public static class DbUtilConfig {
        private boolean writeToDb = false;
        private DbAdapterManager dbAdapterManager;
        private String dbName;

        public boolean isWriteToDb() {
            return writeToDb;
        }

        public void setWriteToDb(boolean writeToDb) {
            this.writeToDb = writeToDb;
        }

        public DbAdapterManager getDbAdapterManager() {
            return dbAdapterManager;
        }

        public void setDbAdapterManager(DbAdapterManager dbAdapterManager) {
            this.dbAdapterManager = dbAdapterManager;
        }

        public String getDbName() {
            return dbName;
        }

        public void setDbName(String dbName) {
            this.dbName = dbName;
        }
    }

    private static DbUtilConfig globalConfig = new DbUtilConfig();

    public static DbUtilConfig getGlobalConfig() {
        return globalConfig;
    }

    /**
     * 根据数据集的数据转化为字符串，无法转化则返回null
     *
     * @param value
     * @return
     */
    public static Object convertDbValue(Object value) {
        try {
            if (null == value) {
                return null;
            }
            String fieldClassName = value.getClass().getName();
            if (value instanceof Clob) {
                Clob clob = (Clob) value;
                return clob.getSubString(1, (int) clob.length());
            }

            if (value instanceof Blob) {
                Blob blob = (Blob) value;
                return blob.getBytes(1, (int) blob.length());
            }

            if ("[B".equals(fieldClassName) || "byte[]".equals(fieldClassName)) {
                byte[] s = (byte[]) value;
                return new String(s);
            }

            return value;
        } catch (Exception ex) {
            throw new DustDbException("转换数据库值失败", null, ex);
        }
    }

    public static void printSql(HSqlCommand cmd) {
        if (!DictGlobalConfig.Instance.isDebug()) {
            return;
        }

        String sql = cmd.getDbExecuteSql((val, idx) -> toRdbValue(val));
        LoggerHelper.info("\r\n" + sql + "\r\n");

    }

    public static void writeLog(String msg, HSqlCommand cmd) {
        if (LOG_TAG.equals(cmd.getTag())) {
            return;
        }
        writeLog(msg, cmd.getDbExecuteSql((val, idx) -> toRdbValue(val)));
    }

    public static void writeLog(String msg, String sql) {
        if (!globalConfig.writeToDb || globalConfig.dbAdapterManager == null) {
            return;
        }

        try {
            IDataAdapter adapter = globalConfig.dbAdapterManager.getAdapter(globalConfig.dbName);
            HSqlCommand insertLog = new HSqlCommand(SQL_LOG);
            insertLog.setParameter("id", SnowFlakeIdWorker.getInstance0().nextId());
            insertLog.setParameter("msg", msg);
            insertLog.setParameter("sql", sql);
            insertLog.setTag(LOG_TAG);
            adapter.update(insertLog);
            adapter.commitAndCloseQuiet();
        } catch (Exception ex) {
            LoggerHelper.debug(ex.getMessage(), ex);
        }
    }

    private static String toRdbValue(Object v) {
        if (v == null) {
            return "null";
        }
        switch (v.getClass().getSimpleName()) {
            case "Integer":
            case "int":
            case "Short":
            case "short":
            case "Long":
            case "long":
            case "Double":
            case "double":
            case "BigDecimal":
            case "Float":
            case "float":
                return TypeUtil.toString(v);
            case "Boolean":
            case "boolean":
            case "byte":
            case "Byte":
                return TypeUtil.toBoolean(v) ? "1" : "0";
            case "Date":
            case "Timestamp":
            case "String":
                return "'" + TypeUtil.toString(v) + "'";
            case "byte[]":
            case "Byte[]":
                return Base64Util.encodeOfBytes((byte[]) v);
            default:
                return StringUtils.EMPTY;
        }
    }

    private static void cloneRow(ResultSet src, HDataRow dest) {
        for (HDataColumn col : dest.getTable().getColumns()) {
            try {
                dest.put(col.getName(), convertDbValue(src.getObject(col.getSrcIndex())));
            } catch (SQLException ex) {
                throw new DustDbException("数据库值读取转换失败", null, ex);
            }
        }
    }

    public static HDataTable toTable(ResultSet rs) {
        HDataTable dt = new HDataTable();
        if (rs == null) {
            return dt;
        }

        try {
            ResultSetMetaData rsmd = rs.getMetaData();
            int columnCount = rsmd.getColumnCount();
            for (int i = 1; i < columnCount + 1; i++) {
                String fieldName = rsmd.getColumnLabel(i);
                if (StringUtils.isEmpty(fieldName) || StringUtils.equalsIgnoreCase(fieldName, "row_num_")) {
                    continue;
                }

                HDataColumn col = new HDataColumn(fieldName);
                String className = rsmd.getColumnClassName(i);
                col.setDataType(className);
                col.setSrcIndex(i);
                dt.getColumns().add(col);
            }

            while (rs.next()) {
                HDataRow dr = dt.addRow();
                cloneRow(rs, dr);
            }

            dt.acceptChanges();
        } catch (SQLException ex) {
            throw new DustDbException("加载数据库数据失败", null, ex);
        }
        return dt;
    }


    public static String getUrl(String dbType, String host, String db, String queryString) {
        host = TypeUtil.toStringEmpty(host);
        db = TypeUtil.toStringEmpty(db);
        queryString = TypeUtil.toStringEmpty(queryString);

        switch (dbType) {
            case DbTypes.Mysql: {
                String url = "jdbc:mysql://" + host + "/" + db;
                if (!StringUtils.isEmpty(queryString)) {
                    url += "?" + queryString;
                }

                return url;
            }
            case DbTypes.PG: {
                String url = "jdbc:postgresql://" + host + "/" + db;
                if (!StringUtils.isEmpty(queryString)) {
                    url += "?" + queryString;
                }

                return url;
            }
            case DbTypes.Oracle:
                return "jdbc:oracle:thin:@//" + host + "/" + db;
            case DbTypes.Mssql:
                return "jdbc:sqlserver://" + host + ";DatabaseName=" + db;
            case DbTypes.ClickHouse: {
                String url = "jdbc:clickhouse://" + host + "/" + db;
                if (!StringUtils.isEmpty(queryString)) {
                    url += "?" + queryString;
                }

                return url;
            }
            case DbTypes.Highgo: {
                String url = "jdbc:highgo://" + host + "/" + db;
                if (!StringUtils.isEmpty(queryString)) {
                    url += "?" + queryString;
                }

                return url;
            }
        }

        throw new IllegalArgumentException("dbType");
    }
}
